#!/usr/bin/env python

import argparse
import logging
import os
import sys

import filelock

try:
    import perfact.loggingtools
except ImportError:
    pass

from .commands.checkout import Checkout
from .commands.execute import Exec
from .commands.fastforward import FF
from .commands.freeze import Freeze
from .commands.layer_init import LayerInit
from .commands.layer_update import LayerUpdate
from .commands.pick import Pick
from .commands.playback import Playback
from .commands.record import Record
from .commands.reformat import Reformat
from .commands.reset import Reset
from .commands.upload import Upload
from .commands.watch import Watch
from .commands.with_lock import WithLock
from .helpers import load_config
from .zodbsync import ZODBSync


class Runner(object):
    """
    Parses arguments to select the correct SubCommand subclass.
    """

    commands = [
        Record,
        Playback,
        Watch,
        Pick,
        Upload,
        WithLock,
        Reset,
        Exec,
        Reformat,
        Checkout,
        Freeze,
        LayerInit,
        LayerUpdate,
        FF,
    ]

    def __init__(self):
        """
        Set up the argument parser with the possible subcommands
        """
        parser = argparse.ArgumentParser(
            description="""
            Tool to sync objects between a ZODB and a git-controlled folder on
            the file system.
        """
        )
        default_configfile = "/etc/perfact/modsync/zodb.py"
        parser.add_argument(
            "--config",
            "-c",
            type=str,
            help="Path to config (default: %s)" % default_configfile,
            default=default_configfile,
        )
        parser.add_argument(
            "--no-lock",
            action="store_true",
            help="Do not acquire lock. Only use inside a with-lock wrapper.",
        )
        if "perfact.loggingtools" in sys.modules:
            perfact.loggingtools.addArgs(parser, name="ZODBSync")

        # Add all available SubCommand classes as sub-command runners, using
        # either the property "subcommand" or the name of the class.
        # The chosen subcommand class will be available as args.command
        subs = parser.add_subparsers()
        for cls in self.commands:
            name = getattr(cls, "subcommand", cls.__name__.lower())
            subparser = subs.add_parser(
                name,
                help=cls.__doc__,
            )
            cls.add_args(subparser)
            subparser.set_defaults(command=cls)

        self.parser = parser

        # These are set by parse()
        self.args = None
        self.logger = None
        self.config = None
        self.lock = None
        self.sync = None
        self.command = None

    def parse(self, *argv):
        """
        Parse the given arguments and set the command accordingly. If no
        arguments are given, sys.argv is used.
        This can be called several times on the same instance to re-use the
        connection, but the config needs to be the same.
        """
        args = self.parser.parse_args(argv if argv else None)
        self.args = args

        if "perfact.loggingtools" in sys.modules:
            logger = perfact.loggingtools.createLogger(args=args, name="ZODBSync")
        else:
            logger = logging.getLogger("ZODBSync")
            logger.setLevel(logging.INFO)
            logger.addHandler(logging.StreamHandler())
            logger.propagate = True

        self.logger = logger
        if getattr(args.command, "use_config", True):
            config = load_config(args.config)
            if self.config is not None and config != self.config:
                self.logger.warning("Reusing runner with different config")
                self.sync = None
            self.config = config

        # Usually, each command needs a connection to the ZODB, but it might
        # explicitly disable it.
        if self.sync is None and getattr(args.command, "connect", True):
            self.sync = ZODBSync(config=self.config, logger=logger)

        if self.config and not args.no_lock:
            self.lock = filelock.FileLock(
                os.path.join(self.config["base_dir"], ".zodbsync.lock")
            )

        self.command = args.command(
            args=self.args,
            logger=self.logger,
            config=self.config,
            sync=self.sync,
            lock=self.lock,
        )

        return self.command

    def run(self, *argv):
        """
        Parse arguments and run command
        """
        self.parse(*argv).run()



#!/usr/bin/env python

from ..subcommand import SubCommand


class Playback(SubCommand):
    """Play back objects from the file system to the Data.FS"""

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "--override",
            "-o",
            action="store_true",
            help="Override object type changes when uploading",
            default=False,
        )
        parser.add_argument(
            "--no-recurse",
            action="store_true",
            help="""Only upload metadata, do not remove elements or recurse.
            Note: If a path no longer present on the file system is given, it
            is still removed.""",
            default=False,
        )
        parser.add_argument(
            "--skip-errors",
            action="store_true",
            help="Skip failed objects and continue",
            default=False,
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Roll back at the end.",
        )
        parser.add_argument(
            "path",
            type=str,
            nargs="*",
            help="Sub-Path in Data.fs to be played back",
        )

    @SubCommand.with_lock
    def run(self):
        self.sync.playback_paths(
            paths=self.args.path,
            recurse=not self.args.no_recurse,
            override=self.args.override,
            skip_errors=self.args.skip_errors,
            dryrun=self.args.dry_run,
        )



import AccessControl.Permission
import zExceptions

"""
Handlers for reading and writing object information that can be implemented in
addition to meta_type specific information
"""


class MixinModObj(object):
    @staticmethod
    def read(obj):  # pragma: no cover
        """
        Read an object and return a dictionary
        """
        raise NotImplementedError

    @staticmethod
    def write(obj, data):  # pragma: no cover
        """
        Write data to an existing object
        """
        raise NotImplementedError

    @staticmethod
    def implements(obj):  # pragma: no cover
        """
        Decide if this handler can be applied to the object in question
        """
        raise NotImplementedError


class AccessControlObj(MixinModObj):
    @staticmethod
    def roles(obj):
        """Read currently set userdefined roles"""
        return tuple(sorted(obj.userdefined_roles()))

    @staticmethod
    def local_roles(obj):
        """Read currently set local roles"""
        # Ignore local owner role if it is trivial
        return list(
            sorted([role for role in obj.get_local_roles() if role[1] != ("Owner",)])
        )

    @staticmethod
    def implements(obj):
        return True

    @staticmethod
    def read(obj):
        result = {}

        is_root = obj.isTopLevelPrincipiaApplicationObject
        if is_root:
            result["is_root"] = True

        roles = AccessControlObj.roles(obj)
        if roles:
            result["roles"] = roles

        local_roles = AccessControlObj.local_roles(obj)
        if local_roles:
            result["local_roles"] = local_roles

        try:
            result["owner"] = obj._owner
        except AttributeError:
            pass

        # The object's settings where they differ from the default (acquire)
        try:
            # we can not use permission_settings() since it only yields
            # permissions for currently valid roles - however, in watch mode,
            # we do not have the acquisition context and therefore might not
            # know all roles, so we need to go one step further down, using
            # Permission.getRoles()
            perm_set = obj.ac_inherited_permissions(1)
            perm_set = [
                AccessControl.Permission.Permission(p[0], p[1], obj) for p in perm_set
            ]
        except AttributeError:
            perm_set = []

        perms = []
        for perm in perm_set:
            roles = perm.getRoles(default=[])
            # for some reason, someone decided to encode whether a permission
            # is acquired by returning either a tuple or a list...
            acquire = isinstance(roles, list)
            roles = list(roles)
            roles.sort()
            if acquire and not len(roles) or is_root and roles == ["Manager"]:
                # Does not deviate from default
                continue
            perms.append((perm.name, acquire, roles))

        if perms:
            perms.sort()
            result["perms"] = perms

        return result

    @staticmethod
    def write(obj, data):

        # Set userdef roles
        cur = AccessControlObj.roles(obj)
        tgt = data.get("roles", [])
        for role in tgt:
            if role not in cur:
                obj._addRole(role)
        todelete = [r for r in cur if r not in tgt]
        if todelete:
            obj._delRoles(todelete)

        # Set local roles
        cur = dict(AccessControlObj.local_roles(obj))
        tgt = dict(data.get("local_roles", tuple()))
        users = set(cur.keys()) | set(tgt.keys())
        for user in users:
            if user not in tgt:
                obj.manage_delLocalRoles([user])
            elif cur.get(user) != tgt[user]:
                obj.manage_setLocalRoles(user, tgt[user])

        # Permission settings
        # permissions that are not stored are understood to be acquired, with
        # no additional roles being granted this permission
        # An exception is the root application object, which can not acquire
        stored_perms = {
            name: (acquire, roles) for name, acquire, roles in data.get("perms", [])
        }
        for role in obj.ac_inherited_permissions(1):
            name = role[0]
            if name in stored_perms:
                roles = stored_perms[name][1]
                if not stored_perms[name][0]:
                    # no acquire, which ist stored in a tuple instead of a list
                    roles = tuple(roles)
            else:
                # the default is to acquire without additional roles - except
                # for the top-level object, where it is not to acquire and
                # allow Manager
                if obj.isTopLevelPrincipiaApplicationObject:
                    roles = ("Manager",)
                else:
                    roles = []
            AccessControl.Permission.Permission(name, [], obj).setRoles(roles)

        # set ownership
        if "owner" in data:
            owner = data["owner"]
            if isinstance(owner, str):
                # backward compatibility for older behavior, where the
                # corresponding UserFolder was not included
                owner = (["acl_users"], owner)

            obj._owner = data["owner"]


class PropertiesObj(MixinModObj):
    @staticmethod
    def implements(obj):
        if hasattr(obj, "aq_explicit"):
            me = obj.aq_explicit
        else:
            me = obj
        return hasattr(me, "propertyMap")

    @staticmethod
    def read(obj):
        props = obj.propertyMap()

        # Optional: Ignore the "title" property if it exists
        props = list([a for a in props if a["id"] != "title"])

        for prop in props:
            prop["value"] = obj.getProperty(prop["id"])
            # Handle inherited properties correctly
            if "mode" in prop:
                val = getattr(obj, prop["id"])
                del prop["mode"]
                prop["value"] = val

        props = [sorted(a.items()) for a in props]

        # Sort the properties
        props.sort()

        if props:
            return {"props": props}

        return {}

    @staticmethod
    def write(obj, data):
        props = [dict(prop) for prop in data.get("props", [])]
        ids = {prop["id"] for prop in props}
        vals = {prop["id"]: prop["value"] for prop in props}
        types = {prop["id"]: prop["type"] for prop in props}
        cur = obj.propertyIds()

        # Delete any property that is superfluous or has the wrong type
        del_ids = [
            p
            for p in cur
            if p != "title" and (p not in ids or types[p] != obj.getPropertyType(p))
        ]
        for p in del_ids:
            try:
                obj.manage_delProperties(ids=[p])
            except zExceptions.BadRequest as e:
                if str(e) == "Cannot delete output_encoding":
                    print("Ignoring failed attempt to delete output_encoding")
                else:
                    raise

        # Add any property that should exist but is missing
        for prop in ids:
            if prop not in cur or prop in del_ids:
                obj._setProperty(prop, vals[prop], types[prop])

        # Change properties that are not deleted and have the wrong value
        chg = {
            p: vals[p]
            for p in ids
            if p not in del_ids
            and p in cur
            and repr(vals[p]) != repr(obj.getProperty(p))
        }
        if chg:
            obj.manage_changeProperties(**chg)


class ZCacheableObj(MixinModObj):
    @staticmethod
    def implements(obj):
        return hasattr(obj, "ZCacheable_getManagerId")

    @staticmethod
    def read(obj):
        meta = {}
        zcachemanager = obj.ZCacheable_getManagerId()
        if zcachemanager:
            meta["zcachemanager"] = zcachemanager
        return meta

    @staticmethod
    def write(obj, data):
        zcachemanager = data.get("zcachemanager", "")
        obj.ZCacheable_setManagerId(zcachemanager)
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import shutil
import subprocess as sp
import sys
import time  # for periodic output

# for "logging in"
import AccessControl.SecurityManagement
import App.config

# for using an explicit transaction manager
import transaction

# For config loading and initial connection, possibly populating an empty ZODB
import Zope2.App.startup
from Zope2.Startup.run import configure_wsgi

from .helpers import (
    StrRepr,
    literal_eval,
    load_config,
    remove_redundant_paths,
    to_string,
)

# Plugins for handling different object types
from .object_types import mod_implemented_handlers, object_handlers

# Monkey patch ZRDB not to connect to databases immediately.
try:
    from Shared.DC.ZRDB import Connection

    Connection.Connection.connect_on_load = False
except ImportError:  # pragma: no cover
    pass


def mod_format(data=None):
    """Make a printable output of the given object data."""

    # This defines which levels of each key should be split into separate lines
    # if they contain an iterable, in addition to the default rule
    rules = {
        "perms": [4],
        "props": [5],
        "local_roles": [4],
    }

    return StrRepr()(data, rules)


def obj_contents(obj):
    """Fetch list of subitems"""
    func = getattr(obj, "objectIds")
    return sorted(func()) if func else []


def mod_read(
    obj=None, onerrorstop=False, default_owner=None, force_default_owner=False
):
    """Build a consistent metadata dictionary for all types."""

    # Known types:
    known_types = list(object_handlers.keys())

    # TODO:
    # - Preconditions ?
    # - Site Access Rules ?

    meta = {}

    # The Zope object type is always in the same place

    meta_type = obj.meta_type
    meta["type"] = meta_type

    # The title should always be readable
    title = getattr(obj, "title", None)
    # see comment in helpers.py:str_repr for why we convert to string
    if isinstance(title, (bytes, str)):
        meta["title"] = to_string(title)

    # Generic and meta type dependent handlers

    if meta_type not in known_types:
        if onerrorstop:
            raise AssertionError(f"Unsupported type: {meta_type}")
        else:
            meta["unsupported"] = meta_type
            return meta

    for handler in mod_implemented_handlers(obj, meta_type):
        meta.update(dict(handler.read(obj)))

    # if default owner is set, remove the owner attribute if it matches the
    # default owner. also when force_default_owner is set
    owner_is_default = meta.get("owner") == (["acl_users"], default_owner)
    if (default_owner) and (owner_is_default or force_default_owner):
        if "owner" in meta:
            del meta["owner"]

    meta["zodbsync_layer"] = getattr(obj, "zodbsync_layer", None)

    return meta


def mod_write(
    data,
    parent=None,
    obj_id=None,
    override=False,
    root=None,
    default_owner=None,
    force_default_owner=False,
    layer=None,
):
    """
    Given object data in <data>, store the object, creating it if it was
    missing. With <override> = True, this method will remove an existing object
    if there is a meta_type mismatch.  If root is given, it should be the
    application root, which is then updated with the metadata in data, ignoring
    parent.
    Returns the existing or created object
    """

    # Retrieve the object meta type.
    d = dict(data)
    meta_type = d["type"]

    no_owner_given = "owner" not in d

    if (default_owner) and (no_owner_given or force_default_owner):
        d["owner"] = (["acl_users"], default_owner)

    if root is None:
        if hasattr(parent, "aq_explicit"):
            obj = getattr(parent.aq_explicit, obj_id, None)
        else:
            obj = getattr(parent, obj_id, None)
    else:
        obj = root

    if obj is not None and not hasattr(obj, "meta_type"):
        logging.getLogger("ZODBSync").warning(
            "Removing property with colliding ID! ({} in {})".format(obj_id, parent)
        )
        parent.manage_delProperties(ids=[obj_id])
        obj = None

    temp_obj = None
    # ID exists? Check for type
    if obj and obj.meta_type != meta_type:
        if not override:
            raise AssertionError(f"Type mismatch for object {data!r}")
        contents = obj_contents(obj)
        if contents:
            # Rename so we can cut+paste the children
            temp_obj = obj
            temp_id = obj_id
            while temp_id in parent.objectIds():
                temp_id += "_"
            parent.manage_renameObject(obj_id, temp_id)
        else:
            # Remove the existing object in override mode
            parent.manage_delObjects(
                ids=[
                    obj_id,
                ]
            )
        obj = None

    # ID is new? Create a minimal object (depending on type)
    if obj is None:
        object_handlers[meta_type].create(parent, data, obj_id)
        if hasattr(parent, "aq_explicit"):
            obj = getattr(parent.aq_explicit, obj_id, None)
        else:
            obj = getattr(parent, obj_id, None)

    # Send an update (depending on type)
    for handler in mod_implemented_handlers(obj, meta_type):
        handler.write(obj, d)

    # Also write zodbsync layer information
    obj.zodbsync_layer = layer

    if temp_obj:
        children = temp_obj.manage_cutObjects(temp_obj.objectIds())
        obj.manage_pasteObjects(children)
        parent.manage_delObjects([temp_id])

    return obj


class ZODBSync:
    """A ZODBSync instance is capable of mirroring a part of the ZODB
    object tree in the file system.

    By default, the syncer creates a subdirectory "__root__" in the
    given directory and can use the methods "record()" and
    "playback()" to get all objects from the ZODB or write them back,
    respectively.
    """

    # We write the binary sources into files ending with appropriate extensions
    # for convenience. This table guesses the most important ones from the
    # "content_type" property.
    content_types = {
        "application/pdf": "pdf",
        "application/json": "json",
        "application/javascript": "js",
        "image/jpeg": "jpg",
        "image/gif": "gif",
        "image/png": "png",
        "text/javascript": "js",
        "text/css": "css",
        "text/html": "html",
        "image/svg+xml": "svg",
    }

    # In some cases, we can deduce the best extension from the object type.
    meta_types = {
        "Z SQL Method": "sql",
        "Script (Python)": "py",
    }

    def __init__(self, config, logger, site="__root__"):
        self.logger = logger
        self.config = config

        self.base_dir = config["base_dir"]
        self.site = site
        self.app_dir = os.path.join(self.base_dir, self.site)
        self.manager_user = config.get("manager_user", "perfact")
        self.default_owner = config.get("default_owner", "perfact")
        self.force_default_owner = config.get("force_default_owner", False)

        # Statistics
        self.num_obj_total = 1
        self.num_obj_current = 0
        self.num_obj_last_report = time.time()

        # Historically, we switched depending on the given config parameter
        # which configure function to call. However, they essentially do the
        # same and depending on the Zope version, only one is available, so it
        # does not matter how the name of the config was given.
        conf_path = self.config.get("wsgi_conf_path")
        if not conf_path:
            conf_path = self.config.get("conf_path")

        # clear arguments to avoid confusing zope configuration procedure
        sys.argv = sys.argv[:1]

        # Read and parse configuration
        configure_wsgi(conf_path)
        # This initially connects to the ZODB (mostly opening a connection to a
        # running ZEO), sets up the application (which, for Zope 2, includes
        # loading any Products provided in the instance) and, if the ZODB
        # happens to be empty, initializes the application and populates it
        # with some default stuff.
        Zope2.App.startup.startup()

        # We do not use the "global" app object, but open a separate connection
        # with a new transaction manager. This ensures that initializing a
        # second ZODBSync object that connects to a different ZEO in the same
        # thread will not yield "client has seen newer transactions than
        # server!" messages (which is mostly relevant for the tests).
        self.tm = transaction.TransactionManager()
        db = App.config.getConfiguration().dbtab.getDatabase("/", is_root=1)
        root = db.open(self.tm).root
        self.app = root.Application

        # Initialize layers
        layerdir = self.config.get("layers", None)
        layers = []
        fnames = []
        if layerdir and os.path.isdir(layerdir):
            fnames = sorted(os.listdir(layerdir))
        for fname in fnames:
            if any([fname.startswith(key) for key in ".~_"]):
                continue
            ident = fname
            if ident.endswith(".py"):
                ident = ident[:-3]
            layer = {
                **{
                    "ident": ident,
                },
                **load_config(f"{layerdir}/{fname}"),
            }
            if "workdir" not in layer or "source" not in layer:
                raise ValueError("Old-style layer config without workdir+source")
            layers.append(layer)
            workdir = layer["workdir"]
            root = f"{workdir}/{site}"
            if not os.path.isdir(root):
                os.makedirs(root, exist_ok=True)
            if not os.path.isdir(f"{workdir}/.git"):
                sp.run(["git", "init"], cwd=workdir, check=True)

        # Append default top-level layer
        layers.append(
            {
                "ident": None,
                "workdir": self.config["base_dir"],
            }
        )
        # Reverse order - index zero is the topmost fallback layer
        self.layers = list(reversed(layers))

        # Make sure the manager user exists
        if self.config.get("create_manager_user", False):
            self.create_manager_user()

    def create_manager_user(self):
        """
        Make sure the manager user exists.
        """
        userfolder = getattr(self.app, "acl_users", None)
        if userfolder is None:
            self.app.manage_addProduct["OFSP"].manage_addUserFolder()
            userfolder = self.app.acl_users
        user = userfolder.getUser(self.manager_user)
        if user is not None:
            return
        self.tm.begin()
        userfolder._doAddUser(self.manager_user, "admin", ["Manager"], [])
        self.logger.warning(
            "Created user %s with password admin because this user does not"
            " exist!" % self.manager_user
        )
        self.tm.commit()

    def start_transaction(self, note=""):
        """Start a transaction with a given note and return the transaction
        manager, so the caller can call commit() or abort()
        """
        # Log in as a manager
        uf = self.app.acl_users
        user = uf.getUser(self.manager_user).__of__(uf)
        if user is None:
            raise AssertionError(
                f"User {self.manager_user} is not available in database."
                " Perhaps you need to set create_manager_user in config.py?"
            )

        self.logger.info("Using user %s" % self.manager_user)
        AccessControl.SecurityManagement.newSecurityManager(None, user)

        self.tm.begin()
        # Set a label for the transaction
        if note:
            self.tm.get().note(note)
        return self.tm

    def source_ext_from_meta(self, meta, obj_id):
        """Guess a good extension from meta data."""

        content_type = None

        # Extract meta data from the key-value list passed.
        meta_type = meta.get("type", None)
        props = meta.get("props", [])
        for prop in props:
            d = dict(prop)
            if d["id"] == "content_type":
                content_type = d["value"]
                break

        # txt is the default extension.
        ext = "txt"
        # If the ID has a period, the extension defaults from the ID.
        if (obj_id or "").find(".") != -1:
            ext = obj_id.rsplit(".", 1)[-1]

        # If there's an extension to use for the object meta_type, use
        # that.
        ext = self.meta_types.get(meta_type, ext)

        # If there's a match in the content_types database, use that.
        ext = self.content_types.get(content_type, ext)
        return ext

    def fs_path(self, path):
        """
        Return filesystem path corresponding to the object path, which might
        start with a /.
        Note that this is not layer-aware and will always return the path in
        the topmost layer.
        """
        return os.path.join(self.app_dir, path.lstrip("/"))

    def fs_pathinfo(self, path):
        """
        Find the correct layer for the object with the given Data.FS path.
        The top (custom) layer may have __deleted__ or __frozen__ markers for
        any of the parents of path which disable looking into lower layers.
        We then find the topmost layer that has a __meta__ file for path.

        __deleted__ has the same consequence as __frozen__ here. Only when
        recording objects and possibly compressing the layers is there a
        difference (if an object reappears and was only marked with
        __deleted__, a compression is possible).

        The children are collected from the subfolders of all remaining layers.

        Input:
        :path: a path in the ZODB like /PerFact/test/

        Return value:
        {
            'path': Original argument
            'fspath': Full path on the filesystem in correct layer, None if
                      object is not present.
            'children': List of effective subobjects
            'layers': self.layers or only the topmost layer if lower layers are
                      masked.
            'layeridx': Index in 'layers' where the object's currently defining
                        representation is found.
        }
        """
        layers = self.layers
        check = self.base_dir
        markers = ["__frozen__", "__deleted__"]
        for part in [self.site] + path.split("/"):
            if not part:
                continue
            check = os.path.join(check, part)
            if not os.path.isdir(check):
                break
            if any([item in markers for item in os.listdir(check)]):
                # Only keep custom layer
                layers = layers[:1]
                break

        result = {
            "path": path,
            "fspath": None,
            "children": [],
            "layers": layers,
            "layeridx": None,
        }
        path = path.lstrip("/")
        candidates = set()  # subfolders on any layer
        children = set()  # those with a __meta__ file on a some layer
        for idx, layer in enumerate(layers):
            fspath = os.path.join(layer["workdir"], self.site, path)
            if not os.path.isdir(fspath):
                continue
            meta = os.path.join(fspath, "__meta__")
            if result["fspath"] is None and os.path.exists(meta):
                result["fspath"] = fspath
                result["layeridx"] = idx

            for entry in os.listdir(fspath):
                if entry in children or entry.startswith("__"):
                    continue
                candidates.add(entry)
                if os.path.exists(os.path.join(fspath, entry, "__meta__")):
                    children.add(entry)
        missing = candidates - children
        if missing:
            raise AssertionError(f"No __meta__ file on any layer: {path}/{missing}")

        result["children"] = sorted(children)
        return result

    def fs_write(self, path, data):
        """
        Write object data out to a folder with the given path.
        """
        # If the custom layer has a __deleted__ marker for this object, remove
        # it.
        base_dir = self.fs_path(path)
        delpath = os.path.join(base_dir, "__deleted__")
        if os.path.exists(delpath):
            os.remove(delpath)

        # Find layer that holds the current version of the object, falling back
        # to the custom layer
        pathinfo = self.fs_pathinfo(path)
        base_dir = pathinfo["fspath"] or base_dir

        # Make directory for the object if it's not already there
        if not os.path.isdir(base_dir):
            self.logger.debug("Will create new directory %s" % path)
            os.makedirs(base_dir)
        old_data = self.fs_read(pathinfo["fspath"])

        # Build object
        exclude_keys = ["source", "zodbsync_layer"]
        meta = {key: value for key, value in data.items() if key not in exclude_keys}
        fmt = mod_format(meta)
        if isinstance(fmt, str):
            fmt = fmt.encode("utf-8")
        source = data.get("source", None)

        new_data = {"meta": fmt.strip()}
        # Only write out sources if unicode or string
        write_source = isinstance(source, (bytes, str))
        src_fname = None
        if write_source:
            # Write bytes or utf-8 encoded text.
            base = "__source__"
            if isinstance(source, str):
                source = source.encode("utf-8")
                base = "__source-utf8__"

            ext = self.source_ext_from_meta(
                meta=meta,
                obj_id=path.rstrip("/").rsplit("/", 1)[-1],
            )
            src_fname = "{}.{}".format(base, ext)
            new_data["src_fnames"] = [src_fname]
            new_data["source"] = source

        if old_data != new_data:
            # Path in top layer, might be different than the one where we read
            # the content
            write_base = self.fs_path(path)
            os.makedirs(write_base, exist_ok=True)

            self.logger.debug("Will write %d bytes of metadata" % len(fmt))
            with open(os.path.join(write_base, "__meta__"), "wb") as f:
                f.write(fmt)

            # Check if there are stray __source* files and remove them first.
            source_files = [
                s
                for s in os.listdir(write_base)
                if s.startswith("__source") and s != src_fname
            ]
            for source_file in source_files:
                os.remove(os.path.join(write_base, source_file))

            if write_source:
                self.logger.debug("Will write %d bytes of source" % len(source))
                with open(os.path.join(write_base, src_fname), "wb") as f:
                    f.write(source)

            # We wrote the object to the topmost layer, so the index where the
            # current representation can be found is zero.
            pathinfo["layeridx"] = 0

        # Compress if possible: Compare object with its representation on disk
        # if the current layer is ignored. If it is the same, remove it in the
        # current layer. Continue with the next layer that holds the object
        for idx, layer in enumerate(pathinfo["layers"]):
            # This is now the layer that we compare the current layer to in
            # order to check if we can compress it.
            if idx <= pathinfo["layeridx"]:
                continue

            fspath = os.path.join(layer["workdir"], self.site, path.lstrip("/"))
            data = self.fs_read(fspath)
            if not data or not data.get("meta"):
                # No representation on this layer
                continue
            if data != new_data:
                # No compression
                break
            # Remove meta file and all source files
            base = os.path.join(
                pathinfo["layers"][pathinfo["layeridx"]]["workdir"],
                self.site,
                path.lstrip("/"),
            )
            os.remove(os.path.join(base, "__meta__"))
            for src in data.get("src_fnames", []):
                os.remove(os.path.join(base, src))
            # Next comparison point
            pathinfo["layeridx"] = idx

        return pathinfo

    def fs_prune(self, pathinfo, contents):
        """
        Remove all subfolders from path that are not in contents.
        Removes the folder from the top-level directory, but if the effective
        folder that defines the object (in a multi-layer setup) still would
        provide it, recreate the directory and add a __deleted__ file.
        """
        relpath = os.path.join(self.site, pathinfo["path"].lstrip("/"))
        base_dir = self.fs_path(pathinfo["path"])
        for item in pathinfo["children"]:
            if item in contents:
                continue
            tgt = os.path.join(base_dir, item)
            if os.path.isdir(tgt):
                self.logger.info("Removing old item %s from filesystem" % item)
                shutil.rmtree(tgt)
            meta = os.path.join(relpath, item, "__meta__")
            # Omit topmost (custom) layer
            for layer in pathinfo["layers"][1:]:
                if not os.path.exists(os.path.join(layer["workdir"], meta)):
                    continue
                # Mask the path as deleted because it is also present
                # in a lower layer
                os.makedirs(tgt, exist_ok=True)
                with open(os.path.join(tgt, "__deleted__"), "wb"):
                    pass
                break

    def fs_prune_empty_dirs(self):
        "Remove all empty directories"
        for layer in self.layers:
            start = os.path.join(layer["workdir"], self.site)
            for root, _, _ in os.walk(start, topdown=False):
                if root == start:
                    continue
                if not os.listdir(root):
                    os.rmdir(root)

    def fs_read(self, fspath):
        """
        Read data from local file system.
        :fspath: is the full filesystem path of the directory.
        Returns a dictionary with
        - the stripped content of the meta file (if there is one)
        - the list of source files if there are any
        - the content of the source file if there is exactly one
        """
        if fspath is None or not os.path.isdir(fspath):
            return {}

        filenames = os.listdir(fspath)
        if "__meta__" not in filenames:
            return {}

        result = {}
        meta_fname = os.path.join(fspath, "__meta__")
        with open(meta_fname, "rb") as f:
            result["meta"] = f.read().strip()

        src_fnames = sorted([a for a in filenames if a.startswith("__source")])
        if src_fnames:
            result["src_fnames"] = src_fnames

        if len(src_fnames) == 1:
            with open(os.path.join(fspath, src_fnames[0]), "rb") as f:
                result["source"] = f.read()
        return result

    def fs_parse(self, fspath, data=None):
        """
        Parse data obtained from fs_read.
        Returns a dictionary with the parsed data from the
        meta file and an additional "source" key.
        Raises an error if there is no meta file or multiple source files
        """
        if data is None:
            data = self.fs_read(fspath)

        if "meta" not in data:
            raise AssertionError(f"Missing meta file: {fspath}")
        src_fnames = data.get("src_fnames", [])
        if len(src_fnames) > 1:
            raise AssertionError(f"Multiple source files in {fspath}")
        result = dict(literal_eval(data["meta"]))
        if src_fnames:
            src_fname = src_fnames[0]
            src = data["source"]
            if src_fname.rsplit(".", 1)[0].endswith("-utf8__"):
                src = src.decode("utf-8")
            result["source"] = src

        return result

    def record(self, paths, recurse=True, skip_errors=False, ignore_removed=False):
        """Record Zope objects from the given paths into the local
        filesystem."""
        # If /a/b as well as /a are to be recorded recursively, drop /a/b
        if recurse:
            remove_redundant_paths(paths)
        for path in paths:
            obj = self.app
            # traverse into the object of interest
            for part in path.split("/"):
                if not part:
                    continue
                if part not in obj.objectIds():
                    # Depending on skip_errors, this yields an error or a
                    # warning later
                    obj = None
                    break
                obj = getattr(obj, part)
            if obj is None and ignore_removed:
                continue
            self.record_obj(obj, path, recurse=recurse, skip_errors=skip_errors)
        self.fs_prune_empty_dirs()

    def record_obj(self, obj, path, recurse=True, skip_errors=False):
        """Record a Zope object into the local filesystem"""
        try:
            data = mod_read(
                obj,
                default_owner=self.default_owner,
                force_default_owner=self.force_default_owner,
            )
        except Exception:
            severity = "Skipping" if skip_errors else "ERROR"
            msg = "{}: Unable to record path {}".format(severity, path)
            if skip_errors:
                self.logger.warning(msg)
                return
            else:
                self.logger.error(msg)
                raise

        pathinfo = self.fs_write(path, data)
        path_layer = pathinfo["layers"][pathinfo["layeridx"]]["ident"]

        current_layer = getattr(obj, "zodbsync_layer", None)
        if current_layer != path_layer:
            with self.tm:
                obj.zodbsync_layer = path_layer

        if not recurse:
            return

        contents = obj_contents(obj) if ("unsupported" not in data) else []
        self.fs_prune(pathinfo, contents)

        # Update statistics
        self.num_obj_total += len(contents)
        now = time.time()
        if now - self.num_obj_last_report > 2:
            self.logger.info(
                "%d obj saved of at least %d, "
                "current path %s" % (self.num_obj_current, self.num_obj_total, path)
            )
            self.num_obj_last_report = now

        for item in contents:
            self.num_obj_current += 1

            child = getattr(obj, item)
            self.record_obj(
                obj=child,
                path=os.path.join(path, item),
                skip_errors=skip_errors,
            )

    def _playback_path(self, pathinfo):
        """
        Play back one object from the file system to the ZODB.

        Params:
            :pathinfo: A dict as returned by fs_pathinfo

        Precondition:
            The parent of the object in question must exist. When a parent as
            well as a child is to be removed, `playback_paths` makes sure to
            call the deletion in the correct order.

        Side effects:
            In addition to the effect on the ZODB, it might add elements to
            `self.playback_todo` and/or `self.playback_fixorder`.
        """
        path = pathinfo["path"]
        if self.recurse:
            self.num_obj_current += 1
            now = time.time()
            if now - self.num_obj_last_report > 2:
                self.logger.info(
                    "%d obj checked of at least %d, current path %s"
                    % (self.num_obj_current, self.num_obj_total, path)
                )
                self.num_obj_last_report = now
        else:
            # be more verbose because every path is explicitly requested
            self.logger.info("Uploading %s" % path)

        # fspath is None if the object is to be deleted
        fs_data = pathinfo["fspath"] and self.fs_parse(pathinfo["fspath"])

        # extend fs_data with layerinfo
        if fs_data:
            fs_data["zodbsync_layer"] = pathinfo["layers"][pathinfo["layeridx"]][
                "ident"
            ]

        # Traverse to the object if it exists
        parent_obj = None
        obj = self.app
        obj_id = None
        obj_path = []
        for part in path.split("/"):
            if not part:
                continue

            if obj is None:
                # Some parent object is missing
                raise ValueError(
                    "Object {} not found when uploading {}".format(
                        "/".join(obj_path), path
                    )
                )

            parent_obj = obj
            obj_id = part
            obj_path.append(part)
            if part in obj.objectIds():
                obj = getattr(obj, part)
            else:
                obj = None

            if obj is None and fs_data is None:
                # Obj does not exist, neither on the file system nor in the
                # Data.FS - nothing to do
                return

        if fs_data is None:
            if obj_id == "acl_users" and path.startswith("/acl_users"):
                return
            self.logger.info("Removing object " + path)
            try:
                parent_obj.manage_delObjects(ids=[obj_id])
            except AttributeError as e:
                msg = (
                    f"\n\nFailed to remove object {path}, "
                    f"original error was {e}.\n"
                    f"Perhaps your layer workdir is empty?\n"
                    f"Possible solution: Execute layer-init or layer-update.\n"
                )
                raise AssertionError(msg) from e
            return

        if "unsupported" in fs_data:
            self.logger.warning("Skipping unsupported object " + path)
            return

        contents = []
        if self.recurse:
            # We remove any to-be-deleted children before updating the object
            # itself, in case a property with the same name is to be created.
            # The addition of new paths is not done here - playback_paths calls
            # them later on.
            contents = pathinfo["children"]
            srv_contents = obj_contents(obj) if obj else []

            # Find IDs in Data.fs object not present in file system
            del_ids = [
                a
                for a in srv_contents
                if a not in contents and not (obj == self.app and a == "acl_users")
            ]
            if del_ids:
                self.logger.warning("Deleting objects " + repr(del_ids))
                obj.manage_delObjects(ids=del_ids)

        try:
            srv_data = (
                dict(
                    mod_read(
                        obj,
                        default_owner=self.manager_user,
                        force_default_owner=self.force_default_owner,
                    )
                )
                if obj is not None
                else None
            )
        except Exception:
            self.logger.exception("Unable to read object at %s" % path)
            raise

        if fs_data != srv_data:
            self.logger.debug("Uploading: %s:%s" % (path, fs_data["type"]))
            try:
                obj = mod_write(
                    fs_data,
                    parent=parent_obj,
                    obj_id=obj_id,
                    override=self.override,
                    root=(obj if parent_obj is None else None),
                    default_owner=self.default_owner,
                    force_default_owner=self.force_default_owner,
                    layer=pathinfo["layers"][pathinfo["layeridx"]]["ident"],
                )
            except Exception:
                # If we do not want to get errors from missing
                # ExternalMethods, this can be used to skip them
                severity = "Skipping" if self.skip_errors else "ERROR"
                msg = "%s %s:%s" % (severity, path, fs_data["type"])
                if self.skip_errors:
                    self.logger.warning(msg)
                    return
                else:
                    self.logger.error(msg)
                    raise

        self.num_obj_total += len(contents)
        if hasattr(object_handlers[fs_data["type"]], "fix_order"):
            # Store the data for later usage by `_playback_fixorder`.
            self.fs_data[path] = fs_data
            self.playback_fixorder.append(path)

        self.playback_todo.extend(
            [
                self.fs_pathinfo("{}{}/".format(path, item))
                for item in reversed(contents)
            ]
        )

    def _playback_fixorder(self, path):
        """
        Fix the order of the subobjects of the object found in path.

        Precondition:
            Called by `playback_paths` and is set up by `_playback_path`, which
            also writes the object data that was read from the FS into
            `self.fs_data`.

        Side effects:
            In addition to the effect on the ZODB, it removes the corresponding
            entry from `self.fs_data`.
        """
        obj = self.app
        for part in path.split("/"):
            obj = getattr(obj, part) if part else obj

        fs_data = self.fs_data[path]
        object_handlers[fs_data["type"]].fix_order(obj, fs_data)
        del self.fs_data[path]

    def prepare_paths(self, paths):
        # normalize paths - cut off filenames and the site name
        paths = {
            (
                path.rsplit("/", 1)[0]
                if (path.rsplit("/", 1)[-1].startswith("__"))
                else path
            )
            for path in paths
        }
        paths = sorted(
            {
                path[len(self.site) :] if path.startswith(self.site) else path
                for path in paths
            }
        )

        if not len(paths):
            return []

        paths = [path.rstrip("/") + "/" for path in paths]
        return paths

    def playback_paths(
        self, paths, recurse=True, override=False, skip_errors=False, dryrun=False
    ):
        self.recurse = recurse
        self.override = override
        self.skip_errors = skip_errors
        paths = self.prepare_paths(paths)

        if recurse:
            paths = remove_redundant_paths(paths)

        self.num_obj_current = 0
        self.num_obj_total = len(paths)

        note = "zodbsync"
        if len(paths) == 1:
            note += ": " + paths[0]
        txn_mgr = self.start_transaction(note=note)

        # Stack of paths that are to be played back (reversed alphabetical
        # order, we pop elements from the top)
        self.playback_todo = []
        todo = self.playback_todo  # local alias

        # Stack of paths for which we need to fix the order after everything
        # below that path has been handled.
        self.playback_fixorder = []
        fixorder = self.playback_fixorder  # local alias

        # Cached read metadata for paths that are in fixorder so we
        # don't need to read it again from disk.
        self.fs_data = {}

        pathinfo = [self.fs_pathinfo(path) for path in paths]
        # Remove all objects that are to be removed so they do not interfere
        # with properties with the same ID that take their place
        lastdel = None
        for entry in pathinfo:
            if lastdel and entry["path"].startswith(lastdel):
                continue
            if entry["fspath"] is not None:
                todo.append(entry)
                continue
            self._playback_path(entry)
            lastdel = entry["path"]
        todo.reverse()

        # Iterate until both stacks are empty. Whenever the topmost element in
        # todo is no longer a subelement of the topmost element of fixorder, we
        # handle the fixorder element, otherwise we handle the todo element.
        try:
            while todo or fixorder:
                if fixorder:
                    # Handle next object on which to fix order unless there are
                    # still subpaths to be handled
                    path = fixorder[-1]
                    if not (todo and todo[-1]["path"].startswith(path)):
                        self._playback_fixorder(fixorder.pop())
                        continue

                entry = todo.pop()
                self._playback_path(entry)
        except Exception:
            self.logger.exception("Error with path: " + entry["path"])
            txn_mgr.abort()
            raise

        if dryrun:
            self.logger.info("Dry-run. Rolling back")
            txn_mgr.abort()
        else:
            txn_mgr.commit()

    def recent_changes(self, since_secs=None, txnid=None, limit=50, search_limit=100):
        """Retrieve all distinct paths which have changed recently. Control how
        far to look back in time by supplying the number of seconds in Unix
        time in "since_secs" or the transaction ID at which to stop scanning in
        "txnid". Retrieves at most "limit" distinct paths.
        """
        paths = []
        newest_txnid = None
        # Clear the request, so we can access undoable_transactions()
        self.app.REQUEST = {}
        # Loop back collecting transactions
        step_size = 10
        cursor = 0
        done = False
        no_records = False
        limit_reached = False
        while cursor < search_limit:
            txns = self.app._p_jar.db().undoInfo(cursor, cursor + step_size)
            if len(txns) == 0 and cursor == 0:
                no_records = True
                break
            for txn in txns:
                if newest_txnid is None:
                    newest_txnid = txn["id"]
                if since_secs and txn["time"] < since_secs:
                    done = True
                    break
                if txnid and txn["id"] == txnid:
                    done = True
                    break
                this_path = txn["description"].split("\n")[0]
                # Ignore transaction descriptions not defining a path
                if not this_path.startswith("/"):
                    continue
                # Cut the method which originated the change, leaving
                # only the object.
                this_path = this_path.rsplit("/", 1)[0]
                if this_path not in paths:
                    paths.append(this_path)
                    if len(paths) >= limit:
                        done = True
                        limit_reached = True
                        break
            if done:
                break
            cursor += step_size
        return {
            "paths": paths,
            "newest_txnid": newest_txnid,
            "no_records": no_records,
            "search_limit_reached": not done,
            "limit_reached": limit_reached,
        }

    def txn_write(self, txnid):
        """Write the newest transaction ID"""
        with open(os.path.join(self.base_dir, "__last_txn__"), "wb") as f:
            f.write(txnid)

    def txn_read(self):
        """Read the newest transaction ID"""
        try:
            with open(os.path.join(self.base_dir, "__last_txn__"), "rb") as f:
                txn = f.read()
        except IOError:
            txn = None
        return txn



import base64
import io
import json
import os
import os.path
import pickle
import random
import shutil
import string
import subprocess
import tempfile
import time
from contextlib import contextmanager

import pytest
import transaction
import ZEO
from AccessControl.SecurityManagement import newSecurityManager

try:
    from unittest import mock
except ImportError:
    import mock

from .. import extedit, helpers, object_types, zodbsync
from ..main import Runner
from . import environment as env


class DummyResponse:
    """
    For mocking the request in extedit test
    """

    def __init__(self, app):
        self.headers = {}
        self.app = app

    def __enter__(self):
        self.orig_request = self.app.REQUEST
        self.app.REQUEST = helpers.Namespace(
            _auth="dummy",
            RESPONSE=self,
        )
        return self

    def __exit__(self, *args):
        self.app.REQUEST = self.orig_request

    def setHeader(self, key, value):
        self.headers[key] = value


class TestSync:
    """
    All tests defined in this class automatically use the environment fixture
    (ZEO, repo etc.)
    """

    @pytest.fixture(scope="class", autouse=True)
    def environment(self, request):
        """
        Fixture that is automatically used by all tests. Initializes
        environment and injects the elements of it into the class.
        """
        myenv = dict(
            zeo=env.ZeoInstance(),
            repo=env.Repository(),
            jslib=env.JSLib(),
        )
        myenv["zopeconfig"] = env.ZopeConfig(zeosock=myenv["zeo"].sockpath())
        myenv["config"] = env.ZODBSyncConfig(env=myenv)

        # inject items into class so methods can use them
        for key, value in myenv.items():
            setattr(request.cls, key, value)

        # Initially record everything and commit it
        self.run("record", "/")
        self.gitrun("add", ".")
        self.gitrun("commit", "-m", "init")
        request.cls.initial_commit = self.get_head_id()

        # at this point, the test is called
        yield

        # clean up items
        for item in myenv.values():
            item.cleanup()

    @pytest.fixture(scope="function", autouse=True)
    def envreset(self, request):
        """
        Reset the environment after each test.
        """
        self.run("record", "/")
        # Call test
        yield
        if getattr(self, "runner", None):
            self.runner.sync.tm.abort()
        cmds = [
            "reset --hard",
            "clean -dfx",
            "checkout autotest",
            "reset --hard {}".format(self.initial_commit),
        ]
        for cmd in cmds:
            self.gitrun(*cmd.split())
        output = self.gitoutput("show-ref", "--heads")
        for line in output.strip().split("\n"):
            commit, refname = line.split()
            refname = refname[len("refs/heads/") :]
            if refname != "autotest":
                self.gitrun("branch", "-D", refname)

        self.run("playback", "--skip-errors", "/")

    @contextmanager
    def newconn(self):
        "Add secondary connection"
        tm = transaction.TransactionManager()
        db = ZEO.DB(self.zeo.sockpath())
        conn = db.open(tm)
        app = conn.root.Application
        with tm:
            # Log in, manage_renameObject checks permissions
            userfolder = app.acl_users
            user = userfolder.getUser("perfact").__of__(userfolder)
            newSecurityManager(None, user)

        yield helpers.Namespace({"tm": tm, "app": app})
        tm.abort()
        conn.close()

    @pytest.fixture(scope="function")
    def conn(self, request):
        """
        Fixture that provides a secondary connection to the same ZEO
        """
        with self.newconn() as conn:
            yield conn

    def mkrunner(self, *cmd):
        """
        Create or update runner for given zodbsync command
        """
        if not hasattr(self, "runner"):
            self.runner = Runner()
        result = self.runner.parse("--config", self.config.path, *cmd)
        self.app = self.runner.sync.app if self.runner.sync else None
        return result

    def run(self, *cmd):
        "Create runner and run"
        self.mkrunner(*cmd).run()

    def gitrun(self, *cmd):
        """
        Run git command.
        """
        subprocess.check_call(["git", "-C", self.repo.path] + list(cmd))

    def gitoutput(self, *cmd):
        """
        Run git command, returning output.
        """
        return subprocess.check_output(
            ["git", "-C", self.repo.path] + list(cmd),
            universal_newlines=True,
        )

    def upload_checks(self, replace_periods=True, ignore=True):
        """A bunch of asserts to call after an upload test has been performed"""
        assert "lib" in self.app.objectIds()
        assert "js" in self.app.lib.objectIds()
        assert "plugins" in self.app.lib.js.objectIds()
        something_js = "something_js" if replace_periods else "something.js"
        assert something_js in self.app.lib.js.plugins.objectIds()
        content = "alert(1);\n"
        data = helpers.to_string(getattr(self.app.lib.js.plugins, something_js).data)
        assert content == data

        assert "css" in self.app.lib.objectIds()
        assert "skins" in self.app.lib.css.objectIds()
        dark_css = "dark_css" if replace_periods else "dark.css"
        assert dark_css in self.app.lib.css.skins.objectIds()
        content = "body { background-color: black; }\n"
        data = helpers.to_string(getattr(self.app.lib.css.skins, dark_css).data)
        assert content == data

        # dont forget ignored files!
        if ignore:
            assert "ignoreme" not in self.app.lib

    def test_record(self):
        """Recorder tests"""
        # Record everything and make sure acl_users exists
        assert os.path.isfile(self.repo.path + "/__root__/acl_users/__meta__")
        # Recording a non-existent object fails
        with pytest.raises(AttributeError):
            self.run("record", "/nonexist")
        # ... unless --skip-errors is given
        self.run("record", "/nonexist", "--skip-errors")
        # Recording with --lasttxn will create the file
        self.run("record", "--lasttxn")
        assert os.path.isfile(os.path.join(self.repo.path, "__last_txn__"))
        # Making a change with a comment indicating the path will make lasttxn
        # pick it up
        tm = self.runner.sync.start_transaction(note="/testpt")
        self.app.manage_addProduct["PageTemplates"].manage_addPageTemplate(
            id="testpt", text="test1"
        )
        tm.commit()
        self.run("record", "--lasttxn")
        assert os.path.isdir(os.path.join(self.repo.path, "__root__/testpt"))

    def test_record_commit(self):
        """Record with --commit (but no mail and no autoreset)"""
        add = self.app.manage_addProduct["PageTemplates"].manage_addPageTemplate
        with self.runner.sync.tm:
            add(id="test", text="test")
        self.run("record", "/", "--commit")
        # Additional run that does no commit since  nothing changed
        self.run("record", "/", "--commit")
        assert os.path.isdir(os.path.join(self.repo.path, "__root__/test"))
        commits = self.gitoutput("log", "--format=%s")
        assert commits == "Generic commit message.\ninit\n"

    def test_record_autoreset(self):
        """Record with --commit --autoreset."""
        add = self.app.manage_addProduct["PageTemplates"].manage_addPageTemplate
        with self.runner.sync.tm:
            add(id="test", text="test")
        self.run("record", "/", "--commit", "--autoreset")
        assert not os.path.isdir(os.path.join(self.repo.path, "__root__/test"))
        commits = self.gitoutput("log", "--format=%s")
        assert commits == "init\n"
        assert "test" not in self.app.objectIds()

    def test_record_unsupported(self):
        """Check that reading /error_log yields an unsupported marker or an
        error."""
        obj = self.runner.sync.app.error_log
        assert "unsupported" in zodbsync.mod_read(obj)
        with pytest.raises(AssertionError):
            zodbsync.mod_read(obj, onerrorstop=True)

    def test_omit_callable_title(self):
        """It omits title attributes which are callable."""
        app = self.app
        obj = app.manage_addProduct["PageTemplates"].manage_addPageTemplate(
            id="test_pt", title="Not-visible", text="test text"
        )

        def patch_title():
            """Callable to test callable titles."""
            return "Show-me"

        # Normal case
        result = zodbsync.mod_read(obj)
        assert "Not-visible" in result["title"]

        # with callable title
        with mock.patch.object(obj, "title", patch_title):
            result = zodbsync.mod_read(obj)
            assert "title" not in result

    def test_playback(self):
        """
        Record everything, change /index_html, play it back and check if the
        contents are correct.
        """
        path = self.repo.path + "/__root__/index_html/__source-utf8__.html"
        content = "<html></html>"
        with open(path, "w") as f:
            f.write(content)
        self.run("playback", "/index_html")
        assert self.app.index_html() == content

    def add_folder(self, name, msg=None, parent=""):
        """
        Add a folder to the root directory and commit it if msg is given
        """
        folder = os.path.join(self.repo.path, "__root__", parent, name)
        os.mkdir(folder)
        with open(folder + "/__meta__", "w") as f:
            f.write(zodbsync.mod_format({"title": "", "type": "Folder"}))
        if msg is not None:
            self.gitrun("add", ".")
            self.gitrun("commit", "-m", msg)

    def get_head_id(self):
        """Return commit ID of current HEAD."""
        return self.gitoutput("show-ref", "--head", "--hash", "HEAD").strip()

    def prepare_pick(self, name="TestFolder", msg="Second commit"):
        """
        Prepare a commit containing a new folder that can be picked onto the
        initialized repository. Returns the commit ID.
        """
        # Add a folder, commit it
        self.add_folder(name, msg)
        commit = self.get_head_id()

        # Reset the commit
        self.gitrun("reset", "--hard", "HEAD~")

        return commit

    def test_pick(self):
        """
        Pick a prepared commit and check that the folder exists.
        """
        commit = self.prepare_pick()
        self.run("pick", commit)

        assert "TestFolder" in self.app.objectIds()

    def test_pick_dryrun(self):
        """
        Pick a prepared commit in dry-run mode and check that the folder does
        not exist.
        """
        commit = self.prepare_pick()
        self.run("pick", commit, "--dry-run")

        assert "TestFolder" not in self.app.objectIds()

    def test_pick_grep(self):
        """
        Prepare three commits where the first and third share a common pattern
        in the commit message, then pick only those.
        """
        msgs = [
            "T123: first commit",
            "T456: second commit",
            "T123: third commit",
        ]
        for nr, msg in enumerate(msgs):
            self.add_folder("Test" + str(nr), msg)
        commit = self.get_head_id()
        self.gitrun("reset", "--hard", "HEAD~3")
        self.run("pick", "--grep=T123", commit)

        ids = self.app.objectIds()
        assert "Test0" in ids
        assert "Test1" not in ids
        assert "Test2" in ids

    def test_pick_range(self):
        """
        Prepare three commits and pick them as a range
        """
        for i in range(3):
            self.add_folder("Test" + str(i), "Commit " + str(i))
        commit = self.get_head_id()
        self.gitrun("reset", "--hard", "HEAD~3")
        self.run("pick", "HEAD.." + commit)
        ids = self.app.objectIds()
        for i in range(3):
            assert "Test" + str(i) in ids

    def test_pick_fail(self):
        """
        Pick a commit twice, making sure it fails and is rolled back.
        Also pick one applyable and one unknown commit.
        """
        commit = self.prepare_pick()
        for second in [commit, "unknown"]:
            with pytest.raises(subprocess.CalledProcessError):
                self.run("pick", commit, second)
            assert "TestFolder" not in self.app.objectIds()
            assert not os.path.isdir(self.repo.path + "/__root__/TestFolder")

    def test_upload_relpath(self):
        """
        Upload JS library from test environment and check for it in Data.fs
        Provide Data.fs path only
        """

        target_jslib_path = self.jslib.path
        target_repo_path = os.path.join("__root__", "lib")

        self.run(
            "upload",
            "--replace-periods",
            "--valid-extensions",
            "css,js",
            target_jslib_path,
            target_repo_path,
        )

        self.upload_checks()

        # we may even omit __root__ in path!
        target_jslib_path = self.jslib.path
        target_repo_path = os.path.join("lib")

        self.run(
            "upload",
            "--replace-periods",
            "--valid-extensions",
            "css,js",
            target_jslib_path,
            target_repo_path,
        )

        self.upload_checks()

        # add another test case showing dot notation also works
        target_jslib_path = self.jslib.path
        target_repo_path = os.path.join(".", "lib")

        self.run(
            "upload",
            "--replace-periods",
            "--valid-extensions",
            "css,js",
            target_jslib_path,
            target_repo_path,
        )

        self.upload_checks()

    def test_upload_options(self):
        """
        Test upload with different options settings.
        """
        target_jslib_path = self.jslib.path
        target_repo_path = os.path.join("__root__", "lib")

        self.run(
            "upload",
            "--valid-extensions",
            "css,js",
            target_jslib_path,
            target_repo_path,
        )
        self.upload_checks(replace_periods=False)

        self.run("upload", "--replace-periods", target_jslib_path, target_repo_path)
        self.upload_checks(ignore=False)

        self.run(
            "upload",
            "--replace-periods",
            "--valid-extensions",
            "  ,,css,js, ",
            target_jslib_path,
            target_repo_path,
        )
        self.upload_checks(ignore=True)

        self.run("upload", target_jslib_path, target_repo_path)
        self.upload_checks(replace_periods=False, ignore=False)

    def test_upload_relpath_fromrepo(self):
        """
        change working directory to repository before upload to simulate
        calling upload from repo leveraging bash path completion
        """
        cur_path = os.getcwd()
        os.chdir(self.repo.path)

        target_jslib_path = self.jslib.path
        target_repo_path = os.path.join(".", "__root__", "lib")

        self.run(
            "upload",
            "--replace-periods",
            "--valid-extensions",
            "css,js",
            target_jslib_path,
            target_repo_path,
        )

        self.upload_checks()

        target_jslib_path = self.jslib.path
        target_repo_path = os.path.join("__root__", "lib")

        self.run(
            "upload",
            "--replace-periods",
            "--valid-extensions",
            "css,js",
            target_jslib_path,
            target_repo_path,
        )

        self.upload_checks()

        os.chdir(cur_path)

    def test_upload_dryrun(self):
        """
        Upload files in dryrun mode, make sure folder is not found in Data.fs
        """
        target_jslib_path = self.jslib.path
        target_repo_path = os.path.join("__root__", "lib")

        self.run(
            "upload",
            "--replace-periods",
            "--valid-extensions",
            "css,js",
            target_jslib_path,
            target_repo_path,
            "--dry-run",
        )

        assert "lib" not in self.app.objectIds()

    def test_emptying_userdefined_roles(self):
        """
        Check fix for #22: if a Folder defines local roles, playback must be
        able to remove them.
        """
        with self.runner.sync.tm:
            self.app._addRole("TestRole")
        self.run("record", "/")
        fname = self.repo.path + "/__root__/__meta__"
        with open(fname, "r") as f:
            lines = f.readlines()
        with open(fname, "w") as f:
            f.writelines([line for line in lines if "TestRole" not in line])
        self.runner.sync.playback_paths(paths=["/"], recurse=False)
        assert self.app.userdefined_roles() == ()

    def test_userdefined_roles_playback(self):
        """
        Test fix #57: Make sure that playback of an object with local roles
        works correctly. Set a local role, record, read out the recording, play
        back, check that it is set correctly, record again and check that the
        recording matches the first one.
        """
        with self.runner.sync.tm:
            self.app._addRole("TestRole")
            self.app.manage_setLocalRoles("perfact", ("TestRole",))
        self.run("record", "/")

        fname = self.repo.path + "/__root__/__meta__"
        with open(fname, "r") as f:
            recording = f.read()
        self.runner.sync.playback_paths(paths=["/"], recurse=False)
        assert self.app.get_local_roles() == (("perfact", ("TestRole",)),)
        self.runner.sync.record("/", recurse=False)
        with open(fname, "r") as f:
            assert recording == f.read()

    def test_addprop(self):
        "Add a property to the root object"
        fname = self.repo.path + "/__root__/__meta__"
        with open(fname, "r") as f:
            content = f.read()
        data = dict(helpers.literal_eval(content))
        prop = {
            "id": "testprop",
            "type": "string",
            "value": "test",
        }
        data["props"] = [list(prop.items())]
        with open(fname, "w") as f:
            f.write(zodbsync.mod_format(data))
        self.run("playback", "/")
        assert self.app.getProperty("testprop") == "test"

    def test_addtokenprop(self):
        "Validate tokens are correctly written"
        fname = self.repo.path + "/__root__/__meta__"
        with open(fname, "r") as f:
            content = f.read()
        data = dict(helpers.literal_eval(content))
        prop = {
            "id": "testprop",
            "type": "tokens",
            "value": ("123", "518"),
        }
        data["props"] = [list(prop.items())]
        with open(fname, "w") as f:
            f.write(zodbsync.mod_format(data))
        self.run("playback", "/")
        assert self.app.getProperty("testprop") == ("123", "518")

    def test_changeprop(self):
        "Change first the value and then the type of a property"
        with self.runner.sync.tm:
            self.app.manage_addProperty("testprop", "test", "string")
        fname = self.repo.path + "/__root__/__meta__"
        self.run("record", "/")
        with open(fname, "r") as f:
            content = f.read()
        data = dict(helpers.literal_eval(content))
        for ptype, pval in [("string", "changed"), ("int", 1)]:
            prop = {
                "id": "testprop",
                "type": ptype,
                "value": pval,
            }
            data["props"] = [list(prop.items())]
            with open(fname, "w") as f:
                f.write(zodbsync.mod_format(data))
            self.run("playback", "/")
            assert self.app.getProperty("testprop") == pval
            assert self.app.getPropertyType("testprop") == ptype

    def test_cacheable(self):
        "Add a RamCacheManager and use it for index_html"
        self.app.manage_addProduct["StandardCacheManagers"].manage_addRAMCacheManager(
            id="http_cache"
        )
        self.app.index_html.ZCacheable_setManagerId("http_cache")
        self.run("record", "/")
        fname = self.repo.path + "/__root__/index_html/__meta__"
        assert "http_cache" in open(fname).read()
        self.run("playback", "/")
        assert self.app.index_html.ZCacheable_getManagerId() == "http_cache"

    def watcher_step_until(self, watcher, cond):
        """
        After we do some changes on the secondary connection for the watcher
        tests, the primary connection might not immediately see the change.
        This helper function checks for a condition with several retries and
        small waiting in between, only failing if the condition keeps being
        false.
        """
        success = False
        for i in range(5):
            watcher.step()
            success = cond()
            if success:
                break
            time.sleep(0.5)
        assert success

    def test_watch_change(self, conn):
        """
        Start the watcher, change something using the second connection without
        commiting yet, do a step on the watcher, make sure the change is not
        yet visible, then commit the change and do another step, making sure
        that it is now present.
        """
        fname = self.repo.path + "/__root__/__meta__"
        watcher = self.mkrunner("watch")
        watcher.setup()
        conn.tm.begin()
        conn.app._addRole("TestRole")
        watcher.step()
        assert "TestRole" not in open(fname).read()
        conn.tm.commit()
        self.watcher_step_until(watcher, lambda: "TestRole" in open(fname).read())

    def test_watch_move(self, conn):
        """
        Create a Page Template, record it using the watcher, rename it and make
        sure the watcher notices. Then add a second one and do a
        three-way-rename in one transaction, making sure the watcher keeps
        track.
        """
        watcher = self.mkrunner("watch")
        watcher.setup()
        root = self.repo.path + "/__root__/"
        src = "/__source-utf8__.html"
        app = conn.app

        add = app.manage_addProduct["PageTemplates"].manage_addPageTemplate
        rename = app.manage_renameObject

        with conn.tm:
            add(id="test1", text="test1")
        self.watcher_step_until(watcher, lambda: os.path.isdir(root + "test1"))

        with conn.tm:
            rename("test1", "test2")
        self.watcher_step_until(watcher, lambda: os.path.isdir(root + "test2"))
        assert not os.path.isdir(root + "test1")

        with conn.tm:
            add(id="test1", text="test2")
        self.watcher_step_until(watcher, lambda: os.path.isdir(root + "test1"))

        assert os.path.isdir(root + "test1")
        assert open(root + "test1" + src).read() == "test2"
        assert open(root + "test2" + src).read() == "test1"

        with conn.tm:
            rename("test1", "tmp")
            rename("test2", "test1")
            rename("tmp", "test2")
        self.watcher_step_until(
            watcher,
            lambda: open(root + "test1" + src).read() == "test1",
        )
        assert open(root + "test1" + src).read() == "test1"
        assert open(root + "test2" + src).read() == "test2"

    def test_watch_dump_setup(self):
        """
        Check output that a spawned initialization subprocess would generate.
        """
        watcher = self.mkrunner("watch")
        watcher.setup()
        stream = io.BytesIO()
        watcher.dump_setup_data(stream=stream)
        data = pickle.loads(stream.getvalue())
        assert set(data.keys()) == {"tree", "txn", "add_oids"}
        tofind = ["/", "/acl_users/", "/index_html/"]
        for obj in data["tree"].values():
            if obj["path"] in tofind:
                tofind.remove(obj["path"])
        assert tofind == []

    def test_ff(self):
        """
        Change the title on a second branch,
        perform a fast-forward merge to it,
        and verify that the change is correctly applied.
        """
        self.gitrun("checkout", "-b", "second")
        path = self.repo.path + "/__root__/index_html/__meta__"
        with open(path) as f:
            lines = f.readlines()
        lines = [
            line if "('title', " not in line else "    ('title', 'test-ff'),\n"
            for line in lines
        ]
        with open(path, "w") as f:
            f.writelines(lines)
        self.gitrun("commit", "-a", "-m", "Change title via ff")

        self.gitrun("checkout", "autotest")
        self.run("ff", "second")
        assert self.app.index_html.title == "test-ff"

    def test_reset(self):
        """
        Change the title of index_html in a second branch, reset to it and
        check that it is played back correctly.
        """
        self.gitrun("checkout", "-b", "second")
        path = self.repo.path + "/__root__/index_html/__meta__"
        with open(path) as f:
            lines = f.readlines()
        lines = [
            line if "('title', " not in line else "    ('title', 'test'),\n"
            for line in lines
        ]
        with open(path, "w") as f:
            f.writelines(lines)
        self.gitrun("commit", "-a", "-m", "Change title")
        self.gitrun("checkout", "autotest")
        self.run("reset", "second")
        assert self.app.index_html.title == "test"

    def test_revert(self):
        """
        Do the same as in test_reset, but afterwards revert it.
        """
        self.test_reset()
        self.run("exec", "git revert HEAD")
        title = self.app.index_html.title
        assert title != "test"

    def test_checkout(self):
        """
        Switch to another branch
        """
        self.run("checkout", "-b", "other")
        # This switches back to autotest, but with a change
        self.test_reset()
        self.run("checkout", "other")
        assert self.app.index_html.title != "test"
        self.run("checkout", "autotest")
        assert self.app.index_html.title == "test"

    def test_exec_checkout(self):
        """
        Prepare two branches and switch between them.
        """
        self.gitrun("branch", "other")
        self.test_reset()
        self.run("exec", "git checkout other")
        title = self.app.index_html.title
        assert title != "test"

    def test_withlock(self):
        "Running with-lock and, inside that, --no-lock, works"
        self.run(
            "with-lock",
            "zodbsync --config {} --no-lock record /".format(self.config.path),
        )

    def test_extedit(self, encoding=None):
        """
        Update /index_html using the external editor launcher
        """
        header_lines = [
            "url: index_html",
            "path: //index_html",
            "auth: dummy",
            "meta-type: Page Template",
            "content-type: text/html",
        ]
        new_source = "test"
        with DummyResponse(self.app) as resp:
            # Read control file
            content = extedit.launch(
                self.app,
                self.app.index_html,
                "/index_html",
            )
            headers, orig_source = content.split("\n\n", 1)
            assert headers == "\n".join(header_lines)
            assert resp.headers["Content-Type"] == ("application/x-perfact-zopeedit")

            # Update to new content
            if encoding:
                orig_source, new_source = [
                    helpers.to_string(base64.b64encode(helpers.to_bytes(item)))
                    for item in [orig_source, new_source]
                ]
            res = extedit.launch(
                self.app,
                self.app.index_html,
                "/index_html",
                source=new_source,
                orig_source=orig_source,
                encoding=encoding,
            )
            assert "success" in res
            assert resp.headers["Content-Type"] == "application/json"
            assert self.app.index_html._text == "test"

            # Try the update again, which must fail because the orig_source no
            # longer matches
            res = extedit.launch(
                self.app,
                self.app.index_html,
                "/index_html",
                source=new_source,
                orig_source=orig_source,
                encoding=encoding,
            )
            assert "error" in json.loads(res)

            # Check for error on invalid path
            res = extedit.launch(
                self.app,
                self.app,
                "/nonexist",
                source="",
                orig_source="",
            )
            assert res == '{"error": "/nonexist not found"}'

    def test_extedit_base64(self):
        self.test_extedit(encoding="base64")

    def test_extedit_binary(self):
        "Test with binary file that is not valid UTF-8"
        self.app.manage_addProduct["OFSP"].manage_addFile(id="blob")
        with DummyResponse(self.app):
            extedit.launch(
                self.app,
                self.app,
                "/blob",
                source=helpers.to_string(base64.b64encode(b"\xff")),
                orig_source="",
                encoding="base64",
            )
            assert self.app.blob.data == b"\xff"

            res = extedit.launch(
                self.app,
                self.app.blob,
                "/blob",
            )
            assert res.endswith("\n\n/w==")

    def meta_file_path(self, *folders):
        """
        takes n folders in order as arguments and returns path to meta file
        """
        path = self.repo.path + "/__root__/"
        for folder in folders:
            path = path + folder + "/"
        path = path + "__meta__"
        return path

    def test_record_structure_and_playback_local_changes(self):
        """
        create structure in zodb and record,
        make local changes in structure, add a local folder, then playback
        and check if changes played back correctly
        """

        # create a structure of folder and sub folder
        folder_1 = "folder_1"
        s_folder_1 = "s_folder_1"
        self.app.manage_addFolder(id=folder_1)
        self.app.folder_1.manage_addFolder(id=s_folder_1, title=s_folder_1)
        assert "s_folder_1" in self.app.folder_1.objectIds()

        # record structure and check that the objects are recorded
        self.run("record", "/")
        assert os.path.isfile(self.meta_file_path(folder_1, s_folder_1))
        # set new title
        path = self.meta_file_path(folder_1, s_folder_1)
        new_title = "new_title"
        content = "[('title', '" + new_title + "'),('type', 'Folder'),]"
        with open(path, "w") as f:
            f.write(content)

        # create metadata for new folder
        new_folder = "new_folder"
        path = (
            self.repo.path
            + "/__root__/"
            + folder_1
            + "/"
            + s_folder_1
            + "/"
            + new_folder
        )
        os.mkdir(path)
        with open(path + "/__meta__", "w") as f:
            f.write(
                """[
                ('id', '{}'),
                ('title', ''),
                ('type', 'Folder'),
            ]""".format(new_folder)
            )

        # playback changes and check if they're existent
        self.run("playback", "/")
        assert new_title == self.app.folder_1.s_folder_1.title
        assert new_folder in self.app.folder_1.s_folder_1.objectIds()

    def test_watch_structure_changes_and_playback_local_changes(self, conn):
        """
        create structure while 'watch' command is running,
        add local changes, then play those changes back and check,
        if those changes played back correctly
        """

        # start watch daemon
        watcher = self.mkrunner("watch")
        watcher.setup()
        app = conn.app
        folder_1 = "folder_1"
        s_folder_1 = "s_folder_1"

        # create folder and wait until watch notices change
        with conn.tm:
            app.manage_addFolder(id=folder_1)
        self.watcher_step_until(
            watcher, lambda: os.path.isdir(self.repo.path + "/__root__/" + folder_1)
        )

        # create subfolder and wait until watch notices change
        with conn.tm:
            app.folder_1.manage_addFolder(id=s_folder_1, title=s_folder_1)
        path = self.repo.path + "/__root__/" + folder_1 + "/" + s_folder_1
        self.watcher_step_until(watcher, lambda: os.path.isdir(path))

        # change title
        new_title = "new_title"
        path = self.meta_file_path(folder_1, s_folder_1)
        content = "[('title', '" + new_title + "'),('type', 'Folder'),]"
        with open(path, "w") as f:
            f.write(content)

        # playback changes and check if those are existent in zodb
        self.run("playback", "/")
        assert new_title == self.app.folder_1.s_folder_1.title

        # wait for watch to notices played back changes
        with open(path) as f:
            meta = f.read()
        self.watcher_step_until(
            watcher, lambda: "('title', '" + new_title + "')" in meta
        )

    def test_watch_structure_changes_and_playback_deleted_folder(self, conn):
        """
        Create structure while 'watch' command is running, remove a folder,
        then play those changes back and check that the watcher handles this
        without crashing.
        """

        # start watch daemon
        watcher = self.mkrunner("watch")
        watcher.setup()
        app = conn.app
        folder_1 = "folder_1"
        s_folder_1 = "s_folder_1"

        # create folder and wait until watch notices change
        with conn.tm:
            app.manage_addFolder(id=folder_1)
        self.watcher_step_until(
            watcher, lambda: os.path.isdir(self.repo.path + "/__root__/" + folder_1)
        )

        # create subfolder and wait until watch notices change
        with conn.tm:
            app.folder_1.manage_addFolder(id=s_folder_1, title=s_folder_1)
        path = self.repo.path + "/__root__/" + folder_1 + "/" + s_folder_1
        self.watcher_step_until(watcher, lambda: os.path.isdir(path))

        # remove folder s_folder_1
        shutil.rmtree(path)

        # playback changes and check if those are existent in zodb
        self.run("playback", "/")

        # wait for watch to notices played back changes
        self.watcher_step_until(watcher, lambda: not os.path.isdir(path))

    def test_commit_on_branch_and_exec_merge(self):
        """
        change to a git feature branch and create a
        structure there, commit it and change back to the autotest branch
        on autotest branch check if changes from feature arent existent,
        then merge feature branch and check if changes have been applied
        correctly
        """

        # change to feature branch and commit created folder/ subfolder
        branch = "feature"
        folder_1 = "folder_1"
        s_folder_1 = "s_folder_1"
        self.run("exec", "git checkout -b {}".format(branch))
        self.app.manage_addFolder(id=folder_1)
        self.app.folder_1.manage_addFolder(id=s_folder_1)
        assert s_folder_1 in self.app.folder_1.objectIds()
        self.run("record", "/")
        assert os.path.isfile(self.meta_file_path(folder_1, s_folder_1))
        self.gitrun("add", "-A")
        self.gitrun("commit", "-m", "test case 3")

        # checkout to autotest and check that changes are not yet existent
        self.run("exec", "git checkout autotest")
        assert not os.path.isfile(self.meta_file_path(folder_1, s_folder_1))
        assert folder_1 not in self.app.objectIds()

        # merge feature branch and check that changes are applied
        self.run("exec", "git merge {}".format(branch))
        assert os.path.isfile(self.meta_file_path(folder_1, s_folder_1))
        assert folder_1 in self.app.objectIds()

    def test_failing_playback_corrupt_metadata(self):
        """
        create a folder in zodb and record it,
        write wrong meta data to the local file system, then playback
        and check if an error occured
        """

        # create new folder and record it
        folder_1 = "folder_1"
        self.app.manage_addFolder(id=folder_1)
        self.run("record", "/")

        # break metadata
        path = self.repo.path + "/__root__/" + folder_1 + "/__meta__"
        content = "[('gandalf', 'ThisIsAWrongKey'),]"
        with open(path, "w") as f:
            f.write(content)

        # test that playback fails
        with pytest.raises(KeyError):
            self.run("playback", "/")

    def test_failing_exec_commands(self):
        """
        call exec commands with wrong commits and
        check if exceptions are thrown correctly
        """
        with pytest.raises(subprocess.CalledProcessError):
            self.run("exec", "revert ThisIsDefinitelyNoCommit")

        with pytest.raises(subprocess.CalledProcessError):
            self.run("exec", "reset ThisIsDefinitelyNoCommit")

        with pytest.raises(subprocess.CalledProcessError):
            self.run("exec", "cherry-pick ThisIsDefinitelyNoCommit")

    def test_create_multiple_commits_on_branch_and_pick_single_on_autotest(self):
        """
        create a feature branch on which
        two changes will be commited to one commit each
        change back to the autotest branch and use pick
        to get the changes of that last commit
        make sure only the last changes are present
        """
        branch = "feature"
        self.gitrun("checkout", "-b", branch)

        # make first changes and commit those
        folder_1 = "folder_1"
        self.app.manage_addFolder(id=folder_1)
        assert folder_1 in self.app.objectIds()
        self.run("record", "/")
        assert os.path.isfile(self.meta_file_path(folder_1))
        self.gitrun("add", "-A")
        self.gitrun("commit", "-m", "pick_commit_1")

        # make second changes and commit those
        folder_2 = "sf_2_tc6"
        self.app.manage_addFolder(id=folder_2)
        assert folder_2 in self.app.objectIds()
        self.run("record", "/")
        assert os.path.isfile(self.meta_file_path(folder_2))
        self.gitrun("add", "-A")
        self.gitrun("commit", "-m", "pick_commit_2")

        commit = self.get_head_id()

        # checkout autotest and check both changes aren't existent
        self.run("exec", "git checkout autotest")
        assert not os.path.isfile(self.meta_file_path(folder_1))
        assert folder_1 not in self.app.objectIds()
        assert not os.path.isfile(self.meta_file_path(folder_2))
        assert folder_2 not in self.app.objectIds()

        # pick 2nd commit and check that
        # first arent' but second changes are applied
        self.run("pick", commit)
        assert not os.path.isfile(self.meta_file_path(folder_1))
        assert folder_1 not in self.app.objectIds()
        assert os.path.isfile(self.meta_file_path(folder_2))
        assert folder_2 in self.app.objectIds()

    def test_create_structure_and_reset_commits(self):
        """
        create structure in zodb and record,
        make local changes in structure, add a local folder,
        commit these changes then playback
        and check if changes played back correctly
        afterwards reset the last comment and check that changes
        are gone
        """

        folder_1 = "folder_1"
        s_folder_1 = "s_folder_1"

        # create first changes and commit those
        self.app.manage_addFolder(id=folder_1)
        self.app.folder_1.manage_addFolder(id=s_folder_1, title=s_folder_1)
        assert s_folder_1 in self.app.folder_1.objectIds()
        self.run("record", "/")
        assert os.path.isfile(self.meta_file_path(folder_1, s_folder_1))

        self.gitrun("add", "-A")
        self.gitrun("commit", "-m", "reset_commit_1")

        # create second changes and commit those
        path = self.repo.path + "/__root__/" + folder_1 + "/" + s_folder_1 + "/__meta__"
        new_title = "new_title"
        content = "[('title', '" + new_title + "'),('type', 'Folder'),]"
        with open(path, "w") as f:
            f.write(content)
        new_folder = "new_folder"
        path = (
            self.repo.path
            + "/__root__/"
            + folder_1
            + "/"
            + s_folder_1
            + "/"
            + new_folder
        )
        os.mkdir(path)
        with open(path + "/__meta__", "w") as f:
            f.write(
                """[
                ('id', '{}'),
                ('title', ''),
                ('type', 'Folder'),
            ]""".format(new_folder)
            )
        self.run("playback", "/")

        self.gitrun("add", "-A")
        self.gitrun("commit", "-m", "reset_commit_2")

        # check that changes are existent in zodb
        assert new_title == self.app.folder_1.s_folder_1.title
        assert new_folder in self.app.folder_1.s_folder_1.objectIds()

        # reset HEAD by one commit and check that second changes are
        # not existent anymore but first changes still are
        self.run("reset", "HEAD~1")
        assert folder_1 in self.app.objectIds()
        assert s_folder_1 in self.app.folder_1.objectIds()
        assert os.path.isfile(self.meta_file_path(folder_1, s_folder_1))
        assert new_title != self.app.folder_1.s_folder_1.title
        assert new_folder not in self.app.folder_1.s_folder_1.objectIds()

        # reset HEAD by one commit and check that first changes are
        # not existent anymore
        self.run("reset", "HEAD~1")
        assert folder_1 not in self.app.objectIds()
        assert not os.path.isfile(self.meta_file_path(folder_1))

    @pytest.mark.parametrize("meta_type", object_types.object_handlers)
    def test_objecttypes(self, meta_type):
        """
        Generic test that is executed for each coded object type. This creates
        an object and writes a modification to it, without actually checking
        for anything. Some are known to fail, for example because they need
        products that are not published on pypi or because they need external
        ressources like non-free libraries for external data connections.
        """
        if meta_type in [
            "DTML TeX",
            "ZForce",
            "External Method",
            "Z cxOracle Database Connection",
            "Z sap Database Connection",
        ]:
            pytest.skip("Skipping objects that require elaborate dependencies")

        if "Test" not in self.app.objectIds():
            self.app.manage_addProduct["OFSP"].manage_addFolder(id="Test")
        if meta_type in ["User Folder", "Simple User Folder"]:
            objid = "acl_users"
        else:
            objid = "testobj"
        parent = self.app.Test
        handler = object_types.object_handlers[meta_type]
        # data that is required by some objects and ignored by others
        add_data = {
            "title": "test",
            "content_type": "text/plain",
            "connection_id": "dbconn",
            "connection_string": "",
            "autocommit": False,
            "maxrows": 100,
            "args": "",
            "source": "",
            "smtp_host": "localhost",
            "smtp_port": "25",
        }
        handler.create(parent, add_data, objid)
        obj = getattr(parent, objid)
        data = zodbsync.mod_read(obj)
        handler.write(obj, data)
        parent.manage_delObjects(ids=[objid])

    def test_ordered_folder_playback(self):
        """
        Checks for the issue recorded in #83: A playback caused by `zodbsync
        exec` that adds a new child to an ordered folder somewhere not at the
        end was still placing it at the end.
        """
        with self.runner.sync.tm:
            self.app.manage_addProduct["OFSP"].manage_addOrderedFolder(id="Test")
            self.app.Test.manage_addProduct["OFSP"].manage_addFolder(id="exist")
        assert self.app.Test.objectIds() == ["exist"]
        self.run("record", "/")

        folder = self.repo.path + "/__root__/Test/"

        new_folder = folder + "new"
        os.mkdir(new_folder)

        with open(os.path.join(new_folder, "__meta__"), "w") as f:
            f.write(
                zodbsync.mod_format(
                    {
                        "title": "",
                        "type": "Folder",
                    }
                )
            )

        with open(folder + "__meta__", "w") as f:
            f.write(
                zodbsync.mod_format(
                    {
                        "contents": ["new", "exist"],
                        "title": "",
                        "type": "Folder (Ordered)",
                    }
                )
            )
        self.run("playback", "--no-recurse", "/Test", "/Test/new")
        assert self.app.Test.objectIds() == ["new", "exist"]

    def test_change_folder_type(self):
        """
        Change a folder to an ordered folder, but without having all children
        in the contents field. The named children must be in the correct order.
        Also check that children are not unnecessarily deleted and recreated by
        a type change.
        Afterwards, change back to Folder and again check that the children
        stay the same.
        Also change the type of a folder without children.
        """

        def add(parent, fid):
            parent.manage_addProduct["OFSP"].manage_addFolder(id=fid)

        with self.runner.sync.tm:
            add(self.app, "Test")
            for child in ["A", "B", "C"]:
                add(self.app.Test, child)
        self.run("record", "/")
        meta = "{}/__root__/Test/__meta__".format(self.repo.path)

        with open(meta, "w") as f:
            f.write(
                zodbsync.mod_format(
                    {
                        "contents": ["B", "A"],
                        "title": "change",
                        "type": "Folder (Ordered)",
                    }
                )
            )
        orig_oid = self.app.Test.A._p_oid
        self.run("playback", "/Test", "--override")
        assert self.app.Test.meta_type == "Folder (Ordered)"
        ids = self.app.Test.objectIds()
        assert sorted(ids) == ["A", "B", "C"]
        assert ids.index("B") < ids.index("A")
        assert self.app.Test.A._p_oid == orig_oid

        with open(meta, "w") as f:
            f.write(
                zodbsync.mod_format(
                    {
                        "title": "change again",
                        "type": "Folder",
                    }
                )
            )
        self.run("playback", "/Test", "--override")
        assert self.app.Test.meta_type == "Folder"
        assert sorted(self.app.Test.objectIds()) == ["A", "B", "C"]
        assert self.app.Test.A._p_oid == orig_oid

        with self.runner.sync.tm:
            self.app.Test.manage_delObjects(ids=["A", "B", "C"])
        self.run("record", "/")
        with open(meta, "w") as f:
            f.write(
                zodbsync.mod_format(
                    {
                        "title": "change",
                        "type": "Folder (Ordered)",
                    }
                )
            )
        self.run("playback", "/Test", "--override")
        assert self.app.Test.meta_type == "Folder (Ordered)"

    def test_create_userfolder(self):
        """
        Check that we can recover from a state where the top-level userfolder
        was deleted.
        Note that we here call create_manager_user manually, but this is not
        necessary when using `zodbsync playback` since it is called upon
        initialization of the `ZODBSync` class instance if the config variable
        is set accordingly. But since the test tries to avoid tearing down and
        recreating the class instance, we need to call it manually.
        """
        with self.runner.sync.tm:
            self.app.manage_delObjects("acl_users")
            self.runner.sync.create_manager_user()
        self.run("playback", "/")
        assert self.app.acl_users.meta_type == "User Folder"

    def test_no_unnecessary_writes(self):
        """
        Check that recording or playing back an unchanged object does not
        actually update it.
        """
        with self.runner.sync.tm:
            self.app.manage_addProduct["OFSP"].manage_addFolder(id="test")

        folder = self.app.test
        mtime1 = folder._p_mtime

        self.run("record", "/test")
        self.run("playback", "/test")
        mtime2 = folder._p_mtime
        assert mtime1 == mtime2

        path = self.repo.path + "/__root__/test/__meta__"
        fsmtime1 = os.stat(path).st_mtime
        self.run("record", "/test")
        fsmtime2 = os.stat(path).st_mtime
        assert fsmtime1 == fsmtime2

    def test_no_meta_file(self):
        """
        Check that a missing meta file regards the object as deleted.
        """

        broken_obj = os.path.join(self.repo.path, "__root__", "foo")
        os.mkdir(broken_obj)

        self.run("playback", "/foo")
        assert "foo" not in self.app.objectIds()

        self.add_folder("Test")
        self.run("playback", "/Test")
        os.remove(os.path.join(self.repo.path, "__root__/Test/__meta__"))
        self.run("playback", "/Test")
        assert "Test" not in self.app.objectIds()

    def test_force_default_owner(self):
        """
        Check if the default owner can be forced via config
        """

        self.runner.sync.force_default_owner = True

        # first test: owner from meta file pushed to app
        folder = os.path.join(self.repo.path, "__root__", "newfolder")
        os.mkdir(folder)

        with open(os.path.join(folder, "__meta__"), "w") as f:
            f.write(
                zodbsync.mod_format(
                    {
                        "title": "",
                        "type": "Folder",
                        "owner": (["acl_users"], "Somebody"),
                    }
                )
            )

        self.run("playback", "/newfolder")

        expected_owner = (["acl_users"], self.runner.sync.default_owner)

        assert self.app.newfolder._owner == expected_owner

        # second test: owner from zope read to meta file
        with self.runner.sync.tm:
            self.app.manage_addProduct["OFSP"].manage_addFolder(id="another")

        self.app.another._owner = (["acl_users"], "Somebody")

        self.run("record", "/")

        meta = self.runner.sync.fs_parse(
            os.path.join(self.repo.path, "__root__/another")
        )

        assert "owner" not in meta

    def test_force_default_owner_negative(self):
        """
        Negative test for force_default_owner setting: Make sure we see
        to old behaviour without this setting being set
        """

        self.runner.sync.force_default_owner = False

        # first test: owner from meta file pushed to app
        folder = os.path.join(self.repo.path, "__root__", "newfolder")
        os.mkdir(folder)

        with open(os.path.join(folder, "__meta__"), "w") as f:
            f.write(
                zodbsync.mod_format(
                    {
                        "title": "",
                        "type": "Folder",
                        "owner": (["acl_users"], "Somebody"),
                    }
                )
            )

        self.run("playback", "/newfolder")
        assert self.app.newfolder._owner == (["acl_users"], "Somebody")

        # second test: owner from zope read to meta file
        with self.runner.sync.tm:
            self.app.manage_addProduct["OFSP"].manage_addFolder(id="another")

        self.app.another._owner = (["acl_users"], "Somebody")

        self.run("record", "/")

        meta = self.runner.sync.fs_parse(
            os.path.join(self.repo.path, "__root__/another")
        )

        assert meta["owner"] == (["acl_users"], "Somebody")

    def test_reformat(self):
        """
        Make a couple of commits with changes to a meta file using the legacy
        format. Then reformat them, checking that no error occurs and that the
        final state uses the new formatting.
        """
        folder = os.path.join(self.repo.path, "__root__/Test")
        os.mkdir(folder)
        fname = os.path.join(folder, "__meta__")

        def commit():
            self.gitrun("add", "__root__/Test/__meta__")
            self.gitrun("commit", "-m", "Test")

        def store(data, strip=False):
            # With strip=False, simulate an older version where there was no
            # newline at the end of meta files
            with open(fname, "w") as f:
                s = helpers.StrRepr()(data, legacy=True)
                if strip:
                    s = s.strip()
                f.write(s)
            commit()

        store(
            {
                "title": "Zope",
                "roles": ["A"],
                "perms": [("View", False, ["Anonymous"])],
            }
        )
        start = self.get_head_id()

        store(
            {
                "title": "Other",
                "roles": ["A", "B"],
                "perms": [("View", True, ["Anonymous", "A"])],
            },
            strip=True,
        )

        # Add a commit that deletes the object while it does not end in a
        # newline. A naive cherry-pick would result in a merge conflict.
        shutil.rmtree(folder)
        commit()
        os.mkdir(folder)

        store(
            {
                "title": "Other",
                "props": [
                    [("id", "columns"), ("type", "tokens"), ("value", ("a", "b", "c"))],
                ],
            }
        )

        self.run("reformat", start)
        with open(fname) as f:
            fmt = f.read()
        assert fmt.strip().split("\n") == [
            "[",
            "    ('props', [",
            "        [('id', 'columns'), ('type', 'tokens'), ('value', (",
            "            'a',",
            "            'b',",
            "            'c',",
            "        ))],",
            "    ]),",
            "    ('title', 'Other'),",
            "]",
        ]

    def test_replace_child_by_property(self):
        """
        Test that it is possible to remove a child and add a property with the
        same name in the same transaction, and also vice versa.
        """
        with self.runner.sync.tm:
            self.app._setProperty("test", "foo", "string")

        self.run("record", "/")
        self.gitrun("add", ".")
        self.gitrun("commit", "-m", "with property")
        c1 = self.get_head_id()

        with self.runner.sync.tm:
            self.app.manage_delProperties(ids=["test"])
            self.app.manage_addProduct["OFSP"].manage_addFolder(id="test")

        self.run("record", "/")
        self.gitrun("add", ".")
        self.gitrun("commit", "-m", "with child")
        c2 = self.get_head_id()

        self.run("reset", c2)
        self.run("reset", c1)

    @contextmanager
    def appendtoconf(self, text):
        """
        Append some text to the config and restore afterward
        """
        with open(self.config.path) as f:
            orig_config = f.read()
        with open(self.config.path, "a") as f:
            f.write("\n" + text + "\n")
        try:
            yield
        finally:
            with open(self.config.path, "w") as f:
                f.write(orig_config)

    def test_playback_postprocess(self):
        """
        Add configuration option for a postprocessing script and check that
        zodbsync reset executes it.
        """
        fname = "{}/postproc".format(self.zeo.path)
        outfile = "{}.out".format(fname)
        script = "\n".join(
            [
                "#!/bin/bash",
                "cat > {}",
            ]
        ).format(outfile)
        with open(fname, "w") as f:
            f.write(script)
        os.chmod(fname, 0o700)
        with self.appendtoconf('run_after_playback = "{}"'.format(fname)):
            self.test_reset()
            with open(outfile) as f:
                assert json.loads(f.read()) == {"paths": ["/index_html/"]}

    def addscript(self, basename, *lines):
        """
        Add executable file into zeo dir, returning the full filename
        """
        fname = "{}/{}".format(self.zeo.path, basename)
        lines = ("#!/bin/bash",) + lines
        with open(fname, "w") as f:
            f.write("\n".join(lines))
        os.chmod(fname, 0o700)
        return fname

    def test_playback_hook(self):
        """
        Add configuration option for a playback hook script and check that
        only the paths returned are played back
        """
        self.add_folder("NewFolder", "First Folder")
        self.add_folder("NewFolder2", "Second Folder")
        commit = self.get_head_id()
        # Reset the commit
        self.gitrun("reset", "--hard", "HEAD~2")

        playback_cmd_out = "{}/playback_cmd.out".format(self.zeo.path)
        playback_cmd = self.addscript(
            "playback_cmd",
            "cat > {}".format(playback_cmd_out),
        )

        playback_hook = self.addscript(
            "playback_hook",
            "echo '{}'".format(
                json.dumps(
                    [
                        {
                            "paths": ["/NewFolder"],
                            "cmd": playback_cmd,
                        }
                    ]
                )
            ),
        )
        with self.appendtoconf('playback_hook = "{}"'.format(playback_hook)):
            self.run("pick", "HEAD..{}".format(commit))

        assert "NewFolder" in self.app.objectIds()
        assert "NewFolder2" not in self.app.objectIds()
        assert os.path.isfile(playback_cmd_out)

    def test_playback_hook_failed(self):
        """
        Add configuration option for a playback hook script with a
        failing cmd and check that all changes are rolled back
        """
        self.add_folder("NewFolder", "First Folder")
        self.add_folder("NewFolder2", "Second Folder")
        commit = self.get_head_id()
        # Reset the commit
        self.gitrun("reset", "--hard", "HEAD~2")

        playback_cmd = self.addscript(
            "playback_cmd",
            "exit 42",
        )
        playback_hook = self.addscript(
            "playback_hook",
            "echo '{}'".format(
                json.dumps(
                    [
                        {
                            "paths": ["/NewFolder"],
                            "cmd": playback_cmd,
                        },
                        {
                            "paths": ["/NewFolder2"],
                        },
                    ]
                )
            ),
        )
        with self.appendtoconf('playback_hook = "{}"'.format(playback_hook)):
            with pytest.raises(AssertionError):
                self.run("pick", "HEAD..{}".format(commit))

            assert "NewFolder" not in self.app.objectIds()
            assert "NewFolder2" not in self.app.objectIds()

    @contextmanager
    def addlayer(self, seqnum="00"):
        """
        Create a temp directory and add a config that uses this as additional
        code layer.
        """
        name = "{}-{}.py".format(
            seqnum, "".join([random.choice(string.ascii_letters) for _ in range(16)])
        )
        path = "{}/layers/{}".format(self.config.folder, name)
        with tempfile.TemporaryDirectory() as layer:
            workdir = f"{layer}/workdir"
            os.makedirs(f"{workdir}/__root__")
            subprocess.run(["git", "init"], cwd=workdir)
            subprocess.run(
                ["git", "config", "user.email", "zodbsync-tester@perfact.de"],
                cwd=workdir,
            )
            subprocess.run(
                ["git", "config", "user.name", "ZODBSync tester"], cwd=workdir
            )
            source = f"{layer}/source"
            os.makedirs(f"{source}/__root__")
            with open(path, "w") as f:
                f.write(f'workdir = "{layer}/workdir"\n')
                f.write(f'source = "{source}"\n')
                f.write(f'ident = "{name}"\n')
            # Force re-reading config
            if hasattr(self, "runner"):
                del self.runner
            try:
                yield layer
            finally:
                if hasattr(self, "runner"):
                    del self.runner
                os.remove(path)

    def test_layer_record_freeze(self):
        """
        Create a folder, copy it into an additional fixed layer, freeze the
        folder and record it. Check that the top layer still has the object
        and a __frozen__ marker.
        """
        self.add_folder("Test", "Test")
        self.run("playback", "/Test")
        with self.addlayer() as layer:
            shutil.copytree(
                "{}/__root__/Test".format(self.repo.path),
                "{}/workdir/__root__/Test".format(layer),
            )
            self.run("freeze", "/")
            self.run("record", "/")
        for fname in ["__meta__", "__frozen__", "Test/__meta__"]:
            assert os.path.exists("{}/__root__/{}".format(self.repo.path, fname))

    def test_layer_record_nofreeze(self):
        """
        Create a folder, copy it into an additional fixed layer and record
        everything. Check that the top layer no longer has the folder.
        """
        self.add_folder("Test", "Test")
        self.run("playback", "/Test")
        with self.addlayer() as layer:
            shutil.copytree(
                "{}/__root__/Test".format(self.repo.path),
                "{}/workdir/__root__/Test".format(layer),
            )
            self.run("record", "/")
        assert not os.path.exists("{}/__root__/Test".format(self.repo.path))

    def test_layer_record_compress_simple(self):
        """
        Test record compression: Create a folder on custom layer,
        then add a new base layer with the same content. Change
        the object to make it fit new base layer, expect object to
        vanish from custom layer after record.
        """

        # in our custom layer we create a folder with title 'Foobar'
        self.add_folder("Test", "Test")
        self.run("playback", "/Test")
        self.app.Test.title = "Foobar"
        self.run("record", "/")

        # ... then we add a new base layer
        with self.addlayer() as layer:
            shutil.copytree(
                "{}/__root__/Test".format(self.repo.path),  # custom layer!
                "{}/workdir/__root__/Test".format(layer),  # new base layer!
            )
            # now create the standard Test folder titled 'Something
            meta = zodbsync.mod_format({"title": "Something", "type": "Folder"})
            with open(f"{layer}/workdir/__root__/Test/__meta__", "w") as f:
                f.write(meta)
            self.run("playback", "/")

            # still 'Foobar' - custom layer wins
            assert self.app.Test.title == "Foobar"

            # now really switch to 'Something' via app
            self.app.Test.title = "Something"

            # ... and record. should remove customized
            # Test folder aka compress
            self.run("record", "/")
            assert not os.path.isdir(os.path.join(self.repo.path, "__root__/Test"))

    @pytest.mark.parametrize("recurse", [True, False])
    def test_layer_playback(self, recurse):
        """
        Set up a base layer, add a path there and play it back.
        """
        self.add_folder("Test")
        with self.addlayer() as layer:
            src = "{}/__root__".format(self.repo.path)
            tgt = "{}/workdir/__root__".format(layer)
            os.rename(src + "/Test", tgt + "/Test")
            cmd = ["playback", "/Test"]
            if not recurse:
                cmd.append("--no-recurse")
            self.run(*cmd)
            assert "Test" in self.app.objectIds()

    def test_layer_playback_frozen_deleted(self):
        """
        Set up a base layer with a folder, but mask it as deleted in the upper
        layer.
        """
        self.add_folder("Test")
        with self.addlayer() as layer:
            src = "{}/__root__".format(self.repo.path)
            tgt = "{}/workdir/__root__".format(layer)
            shutil.copytree(src + "/Test", tgt + "/Test")
            with open("{}/__frozen__".format(src), "w"):
                pass
            os.remove(src + "/Test/__meta__")
            self.run("playback", "/Test")
            assert "Test" not in self.app.objectIds()

    def test_layer_playback_combined(self):
        """
        Set up a complex hierarchy with two layers and one path being frozen
        and providing different subobjects, one path being merged while also
        changing the object itself and one path being merged without changing
        the object itself.
        """
        for folder in ["Test1", "Test2", "Test3"]:
            self.add_folder(folder)
            for sub in ["Sub1", "Sub2"]:
                self.add_folder(sub, parent=folder)
        with self.addlayer() as layer:
            root = os.path.join(self.repo.path, "__root__")
            # Move current structure into lower layer
            os.rename(root, os.path.join(layer, "workdir/__root__"))
            # Create a sparse structure in top layer
            files = [
                "Test1/__frozen__",
                "Test1/__meta__",
                "Test1/Sub3/__meta__",
                "Test2/__meta__",
                "Test2/Sub3/__meta__",
                "Test3/Sub3/__meta__",
            ]
            meta = """[
                ('props', []),
                ('title', 'overwritten'),
                ('type', 'Folder'),
            ]"""
            for file in files:
                dirname, fname = file.rsplit("/", 1)
                os.makedirs(os.path.join(root, dirname), exist_ok=True)
                with open(os.path.join(root, file), "w") as f:
                    if fname == "__meta__":
                        f.write(meta)

            self.run("playback", "/")
        assert self.app.Test1.objectIds() == ["Sub3"]
        assert self.app.Test2.objectIds() == ["Sub1", "Sub2", "Sub3"]
        assert self.app.Test3.objectIds() == ["Sub1", "Sub2", "Sub3"]
        assert self.app.Test2.title == "overwritten"
        assert self.app.Test3.title == ""

    def test_layer_record(self):
        """
        Add an object and move it to the lower layer. Record again. The object
        must not be added to the top layer since it is already present in the
        lower layer.
        """
        self.add_folder("Test")
        self.run("playback", "/Test")
        self.run("record", "/Test")
        with self.addlayer() as layer:
            root = [
                os.path.join(layer, "workdir/__root__"),
                os.path.join(self.repo.path, "__root__"),
            ]
            os.rename(os.path.join(root[1], "Test"), os.path.join(root[0], "Test"))
            self.run("record", "/Test")
            assert not os.path.isdir(os.path.join(root[1], "Test"))

    def test_layer_record_deletion(self):
        """
        Have an object with subobjects defined in the lower layer, but not in
        the Data.FS. Record it. The top-level layer needs to recreate the
        folder and mark it as deleted.
        """
        self.add_folder("Test")
        self.add_folder("Sub", parent="Test")
        with self.addlayer() as layer:
            srcroot = os.path.join(self.repo.path, "__root__")
            tgtroot = os.path.join(layer, "workdir/__root__")
            os.rename(os.path.join(srcroot, "Test"), os.path.join(tgtroot, "Test"))
            self.run("record", "/")
            assert os.path.isdir(os.path.join(srcroot, "Test"))
            assert os.path.exists(os.path.join(srcroot, "Test/__deleted__"))

    def test_layer_record_prune(self):
        """
        Use a setup with two layers. Add a folder and record it to the custom
        layer. Remove the folder and record again - check that the subfolder is
        actually deleted and not marked with __deleted__.
        """
        self.app.manage_addFolder(id="Test")
        self.run("record", "/")
        with self.addlayer() as layer:
            os.rename(
                os.path.join(self.repo.path, "__root__/__meta__"),
                os.path.join(layer, "workdir/__root__/__meta__"),
            )
            self.run("record", "/")
        assert not os.path.isdir(os.path.join(self.repo.path, "__root__/Test"))

    def test_layer_watch_rename(self):
        """
        Rename an object in the Data.FS that is recorded in a lower layer.
        Check that the watcher does the right thing, marking the original
        object as deleted and creating the new object.
        """
        with self.addlayer() as layer:
            os.rename(
                os.path.join(self.repo.path, "__root__/index_html"),
                os.path.join(layer, "workdir/__root__/index_html"),
            )
            watcher = self.mkrunner("watch")
            watcher.setup()

            # Somehow, we need to initialize the connection here and can not
            # use the fixture, otherwise we are not logged in (probably some
            # interference with addlayer resetting the original connection)
            with self.newconn() as conn:
                with conn.tm:
                    conn.app.manage_renameObject("index_html", "something")
            watcher.step()
            assert os.path.exists(
                os.path.join(self.repo.path, "__root__/index_html/__deleted__")
            )
            assert os.path.exists(
                os.path.join(self.repo.path, "__root__/something/__meta__")
            )

    def test_layer_watch_paste(self):
        """
        Set up two folders, where one has a subfolder, both in the lower layer.
        Cut the subfolder and paste it into the other folder, checking the
        result. Then cut it again and paste it into its original place and
        check that.
        """
        with self.runner.sync.tm:
            self.app.manage_addFolder(id="Test1")
            self.app.manage_addFolder(id="Test2")
            self.app.Test1.manage_addFolder(id="Sub")
        self.run("record", "/")
        with self.addlayer() as layer:
            src = os.path.join(self.repo.path, "__root__")
            tgt = os.path.join(layer, "workdir/__root__")
            os.rmdir(tgt)
            os.rename(src, tgt)
            os.mkdir(src)
            watcher = self.mkrunner("watch")
            watcher.setup()
            with self.newconn() as conn:
                with conn.tm:
                    cp = conn.app.Test1.manage_cutObjects(["Sub"])
                    conn.app.Test2._pasteObjects(cp)
            paths = [
                os.path.join(self.repo.path, "__root__", path)
                for path in ["Test1/Sub/__deleted__", "Test2/Sub/__meta__"]
            ]
            self.watcher_step_until(watcher, lambda: all(map(os.path.exists, paths)))
            with self.newconn() as conn:
                with conn.tm:
                    cp = conn.app.Test2.manage_cutObjects(["Sub"])
                    conn.app.Test1._pasteObjects(cp)

            paths = [
                os.path.join(self.repo.path, "__root__", path)
                for path in ["Test1", "Test2"]
            ]
            # Both folders must be removed
            self.watcher_step_until(watcher, lambda: not any(map(os.path.isdir, paths)))

    def test_layer_recreate_deleted(self):
        """
        Delete an object from the custom layer s.t. it obtains a __deleted__
        marker. Recreate it and make sure that it is no longer present in the
        custom layer since it is the same as below.
        """
        with self.runner.sync.tm:
            self.app.manage_addFolder(id="Test")

        with self.addlayer() as layer:
            self.run("record", "/Test")
            root = os.path.join(self.repo.path, "__root__")
            os.rename(
                os.path.join(root, "Test"),
                os.path.join(layer, "workdir/__root__/Test"),
            )
            self.app.manage_delObjects(ids=["Test"])
            self.run("record", "/")
            assert os.path.exists(os.path.join(root, "Test/__deleted__"))
            self.app.manage_addFolder(id="Test")
            self.run("record", "/Test")
            assert not os.path.isdir(os.path.join(root, "Test"))

    def test_layer_remove_subfolder(self):
        """
        Set up a folder with a subfolder, both only defined in the lower layer.
        Remove the subfolder. Check that both folders are created in the custom
        folder, without __meta__ but in order to correctly place the
        __deleted__ marker.
        """
        with self.runner.sync.tm:
            self.app.manage_addFolder(id="Test")
            self.app.Test.manage_addFolder(id="Sub")

        with self.addlayer() as layer:
            self.run("record", "/")
            root = os.path.join(self.repo.path, "__root__")
            os.rename(
                os.path.join(root, "Test"),
                os.path.join(layer, "workdir/__root__/Test"),
            )
            with self.runner.sync.tm:
                self.app.Test.manage_delObjects(ids=["Sub"])
            self.run("record", "/")
            assert not os.path.exists(os.path.join(root, "Test/__meta__"))
            assert not os.path.exists(os.path.join(root, "Test/Sub/__meta__"))
            assert os.path.exists(os.path.join(root, "Test/Sub/__deleted__"))

    def test_layer_update(self, caplog):
        """
        Set up a layer, and register it. Change something in the layer and use
        layer-update to play back the changed object.
        """
        with self.runner.sync.tm:
            self.app.manage_addFolder(id="Test")
        with self.addlayer() as layer:
            self.run("record", "/")
            ident = self.runner.sync.layers[-1]["ident"]
            src = os.path.join(self.repo.path, "__root__")
            tgt = os.path.join(layer, "source/__root__")
            os.rmdir(tgt)
            os.rename(src, tgt)
            os.mkdir(src)
            self.run("layer-init", "*")
            with open(os.path.join(tgt, "Test/__meta__"), "w") as f:
                f.write(zodbsync.mod_format({"title": "Changed", "type": "Folder"}))
            self.run("layer-update", ident)
            assert "Conflict with object" not in caplog.text
            assert self.app.Test.title == "Changed"

    def test_keep_acl(self):
        """
        Make sure deletions on top level acl_users are NOT synced into
        Data.fs

        User folders living somewhere else in the application may be
        deleted though.
        """
        acl_path = os.path.join(
            self.repo.path,
            "__root__",
            "acl_users",
        )
        shutil.rmtree(acl_path)
        self.run("playback", "/")

        # this playback will fail horribly if acl_users is gone!
        self.run("playback", "/")

        # make sure acl_users in toplevel is still present
        assert "acl_users" in self.app.objectIds()

        # now create dummy module with its own acl_users folder
        with self.runner.sync.tm:
            self.app.manage_addFolder(id="some_module")
            self.app.some_module.manage_addUserFolder()

        self.run("record", "/")

        assert "acl_users" in self.app.some_module.objectIds()

        module_acl = os.path.join(
            self.repo.path,
            "__root__",
            "some_module",
            "acl_users",
        )
        shutil.rmtree(module_acl)
        self.run("playback", "/")
        assert "acl_users" not in self.app.some_module.objectIds()

    def test_keep_acl_norecurse(self):
        """
        test_keep_acl but slightly altered for norecurse,
        aka playing back single objects instead of the whole
        object tree
        """
        acl_path = os.path.join(
            self.repo.path,
            "__root__",
            "acl_users",
        )
        shutil.rmtree(acl_path)
        self.run("playback", "--no-recurse", "/acl_users")

        # make sure acl_users in toplevel is still present
        assert "acl_users" in self.app.objectIds()

        # now create dummy module with its own acl_users folder
        with self.runner.sync.tm:
            self.app.manage_addFolder(id="some_module")
            self.app.some_module.manage_addFolder(id="something")
            self.app.some_module.manage_addUserFolder()

        self.run("record", "/")

        assert "acl_users" in self.app.some_module.objectIds()

        module_acl = os.path.join(
            self.repo.path,
            "__root__",
            "some_module",
            "acl_users",
        )
        shutil.rmtree(module_acl)
        self.run(
            "playback",
            "--no-recurse",
            "/some_module",
            "/some_module/acl_users",
        )
        assert "acl_users" not in self.app.some_module.objectIds()

    def test_layer_update_warn(self, caplog):
        """
        Set up a layer and initialize it. Change an object that is provided by
        this layer and record the change into the custom layer. Update the base
        layer such that this object would change and make sure that we are
        warned that the change is ignored due to a collision.
        Also check that deletion of an object in the base layer that is not
        masked in the custom layer, but that has a masked subobject, also leads
        to a warning.
        """
        with self.runner.sync.tm:
            self.app.manage_addFolder(id="Test")
            self.app.manage_addFolder(id="ToDelete")
            self.app.ToDelete.manage_addFolder(id="Sub")
        with self.addlayer() as layer:
            self.run("record", "/")
            ident = self.runner.sync.layers[-1]["ident"]
            src = os.path.join(self.repo.path, "__root__")
            tgt = os.path.join(layer, "source/__root__")
            os.rmdir(tgt)
            os.rename(src, tgt)
            os.mkdir(src)
            self.run("layer-init", "*")
            with self.runner.sync.tm:
                self.app.Test._setProperty("nav_hidden", True, "boolean")
                self.app.ToDelete.Sub._setProperty("nav_hidden", True, "boolean")
            self.run("record", "/")
            with open(os.path.join(tgt, "Test/__meta__"), "w") as f:
                f.write(zodbsync.mod_format({"title": "Changed", "type": "Folder"}))
            shutil.rmtree(os.path.join(tgt, "ToDelete"))
            self.run("layer-update", ident)
            expect = "Conflict with object in custom layer: "
            assert expect + "/Test" in caplog.text
            assert "AttributeError" not in caplog.text
            assert expect + "/ToDelete/Sub" in caplog.text

    def test_layer_change_into_top(self):
        """
        Verify that changed files are written into the top layer.
        Note that this is not what we want in the long run, but until we have
        methods for moving objects between layers and there is a frontend for
        showing unstaged changes in all layers, everything is written into the
        top layer.
        """
        with self.runner.sync.tm:
            self.app.manage_addProduct["OFSP"].manage_addFile(id="blob")

        with self.addlayer() as layer:
            self.run("record", "/blob")
            shutil.move(
                "{}/__root__/blob".format(self.repo.path),
                "{}/workdir/__root__/blob".format(layer),
            )
            with self.runner.sync.tm:
                self.app.blob.manage_edit(
                    filedata="text_content", content_type="text/plain", title="BLOB"
                )
            self.run("record", "/")
            root = os.path.join(self.repo.path, "__root__")
            # both meta and source file are in custom layer
            assert os.path.exists(os.path.join(root, "blob/__meta__"))
            assert os.path.exists(os.path.join(root, "blob/__source__.txt"))
            source_fmt = "{}/__root__/blob/__source__.txt"
            with open(source_fmt.format(f"{layer}/workdir")) as f:
                # source in layer should still be empty
                assert f.read() == ""
            with open(source_fmt.format(self.repo.path)) as f:
                # ... content is in custom layer!
                assert f.read() == "text_content"

    def test_layer_playback_hook(self):
        """
        Set up two layers. Pick a commit that marks an object as __deleted__ in
        the top layer. Check that the playback hook script gets the normalized
        object paths and not the specific files.
        """
        with self.runner.sync.tm:
            self.app.manage_addProduct["OFSP"].manage_addFile(id="blob")

        root = "{}/__root__".format(self.repo.path)
        with self.addlayer() as layer:
            self.run("record", "/blob")
            shutil.move(
                "{}/blob".format(root),
                "{}/workdir/__root__/blob".format(layer),
            )
            os.mkdir("{}/blob".format(root))
            with open("{}/blob/__deleted__".format(root), "w"):
                pass
            self.gitrun("add", ".")
            self.gitrun("commit", "-m", "delete blob")
            commid = self.get_head_id()
            self.gitrun("reset", "--hard", "HEAD~")
            output = "{}/playback_hook.out".format(self.zeo.path)
            playback_hook = self.addscript(
                "playback_hook",
                "cat > {}".format(output),
                "echo '[]'",
            )
            with self.appendtoconf('playback_hook = "{}"'.format(playback_hook)):
                self.run("pick", commid)
            with open(output) as f:
                assert {"paths": ["/blob/"]} == json.loads(f.read())

    def test_layer_tar(self):
        """
        Perform a layer-init and layer-update from a tar file source.
        """
        with self.runner.sync.tm:
            self.app.manage_addProduct["OFSP"].manage_addFile(id="blob")
        with self.addlayer() as layer:
            self.run("record", "/blob")
            subprocess.run(
                ["tar", "cf", f"{layer}/source/__root__.tar", "blob"],
                cwd=f"{self.repo.path}/__root__",
                check=True,
            )
            os.rmdir(f"{layer}/source/__root__")
            self.run("layer-init", "*")
            assert os.listdir(f"{layer}/workdir/__root__") == ["blob"]
            # Record to remove from fallback layer
            self.run("record", "/")
            assert "blob" not in os.listdir(f"{self.repo.path}/__root__")
            # Now change the file in the TAR file and run layer-update
            shutil.copytree(
                f"{layer}/workdir/__root__/blob",
                f"{layer}/blob",
            )
            with open(f"{layer}/blob/__source__.txt", "w") as f:
                f.write("changed")
            subprocess.run(
                ["tar", "cf", f"{layer}/source/__root__.tar", "blob"],
                cwd=layer,
                check=True,
            )
            self.run("layer-update", "*")
            assert str(self.app.blob) == "changed"

    def test_layer_update_2phase_failed(self):
        """
        Perform layer-update with a two-phase playback where a command at the
        end of the first phase fails. Check that the rollback is performed
        correctly.
        """
        with self.runner.sync.tm:
            self.app.manage_addProduct["OFSP"].manage_addFile(id="blob")
        playback_cmd = self.addscript("playback_cmd", "false")

        playback_hook = self.addscript(
            "playback_hook",
            "echo '{}'".format(
                json.dumps(
                    [
                        {
                            "paths": ["/blob"],
                            "cmd": playback_cmd,
                        }
                    ]
                )
            ),
        )
        with self.appendtoconf('playback_hook = "{}"'.format(playback_hook)):
            with self.addlayer() as layer:
                self.run("record", "/")
                src = f"{self.repo.path}/__root__/blob"
                tgt = f"{layer}/source/__root__/blob"
                os.rename(src, tgt)
                self.run("layer-init", "*")
                with open(f"{tgt}/__source__.txt", "w") as f:
                    f.write("changed")
                with pytest.raises(AssertionError):
                    self.run("layer-update", "*")
                assert str(self.app.blob) == ""

    def test_layer_info_datafs(self):
        """
        Validate the correct writing and clearing of the layer ident
        in the Data.FS
        """
        with self.runner.sync.tm:
            self.app.manage_addProduct["OFSP"].manage_addFile(id="blob")

        with self.addlayer() as layer:
            self.run("record", "/blob")
            assert getattr(self.app.blob, "zodbsync_layer", None) is None
            # Move file to layer and check that layer info is stored in Data.FS
            shutil.move(
                "{}/__root__/blob".format(self.repo.path),
                "{}/workdir/__root__/blob".format(layer),
            )
            self.run("record", "/")
            assert getattr(self.app.blob, "zodbsync_layer") is not None
            # Change file in Data.FS and verify that layer info is cleared
            with self.runner.sync.tm:
                self.app.blob.manage_edit(
                    filedata="text_content", content_type="text/plain", title="BLOB"
                )
            self.run("record", "/")
            assert getattr(self.app.blob, "zodbsync_layer", None) is None

    def test_fail_when_meta_is_missing(self):
        """
        Check that playing back a structure where no layer has a meta file for
        a given folder does not work.
        """
        root = f"{self.repo.path}/__root__"
        os.mkdir(f"{root}/newfolder")
        os.mkdir(f"{root}/newobj")
        with open(f"{root}/newobj/__source__.py", "w"):
            pass
        with pytest.raises(AssertionError):
            self.run("playback", "/")

    def test_fail_when_meta_missing_layers(self):
        """
        Check that playing back a structure where no layer has a meta file for
        a given folder does not work (multi-layer).
        """
        with self.addlayer() as layer:
            os.mkdir(f"{self.repo.path}/__root__/newfolder")
            os.mkdir(f"{layer}/workdir/__root__/newfolder")
            with pytest.raises(AssertionError):
                self.run("playback", "/")



import os
import shutil
import subprocess
import tempfile

"""
These classes together define an environment containing of a running ZEO
instance and a git repository (both running in temporary directories) as well
as a Zope instance configuration connecting to this ZEO and a ZODBSync
configuration connecting everything. They are used by the fixture defined in
conftest.py to provide an environment for the tests.
"""


class ZeoInstance:
    def __init__(self):
        self.path = tempfile.mkdtemp()
        subprocess.check_call(["mkzeoinstance", self.path])

        # replace address line to use a socket
        fname = self.path + "/etc/zeo.conf"
        with open(fname) as f:
            lines = f.readlines()
        subst = "  address " + self.sockpath() + "\n"
        lines = [subst if "  address" in line else line for line in lines]
        with open(fname, "w") as f:
            f.writelines(lines)

        self.zeo = subprocess.Popen([self.path + "/bin/runzeo"])

    def sockpath(self):
        return self.path + "/var/zeo.sock"

    def cleanup(self):
        self.zeo.terminate()
        self.zeo.wait()
        shutil.rmtree(self.path)


class Repository:
    def __init__(self):
        self.path = tempfile.mkdtemp()
        commands = [
            ["init"],
            ["branch", "-m", "autotest"],
            ["config", "user.email", "test@zodbsync.org"],
            ["config", "user.name", "testrepo"],
        ]
        for cmd in commands:
            subprocess.check_call(["git", "-C", self.path] + cmd)

    def cleanup(self):
        shutil.rmtree(self.path)


class ZopeConfig:
    def __init__(self, zeosock):
        self.path = tempfile.mkdtemp()
        self.config = self.path + "/zope.conf"
        content = """
%define INSTANCE {path}
%define ZEO_SERVER {zeosock}

instancehome $INSTANCE

<zodb_db main>
    <zeoclient>
      server $ZEO_SERVER
      storage 1
      name zeostorage
      var $INSTANCE/var
      cache-size 20MB
    </zeoclient>
   mount-point /
</zodb_db>
        """.format(zeosock=zeosock, path=self.path)

        with open(self.config, "w") as f:
            f.write(content)

    def cleanup(self):
        shutil.rmtree(self.path)


class ZODBSyncConfig:
    def __init__(self, env):
        self.folder = tempfile.mkdtemp()
        os.mkdir(self.folder + "/layers")
        self.path = self.folder + "/zodb.py"
        with open(self.path, "w") as f:
            f.write(
                """
conf_path = '{zopeconf}'
datafs_path = '{zeopath}/var/Data.fs'
manager_user = 'perfact'
create_manager_user = True
default_owner = 'perfact'
base_dir = '{repodir}'
commit_name = "Zope Developer"
commit_email = "zope-devel@example.de"
commit_message = "Generic commit message."
layers = "{root}/layers"
            """.format(
                    zopeconf=env["zopeconfig"].config,
                    zeopath=env["zeo"].path,
                    repodir=env["repo"].path,
                    root=self.folder,
                )
            )

    def cleanup(self):
        shutil.rmtree(self.folder)


class JSLib:
    """
    A test JS library containing some JS and CSS files
    """

    def __init__(self):
        self.path = tempfile.mkdtemp()

        self.js_folder = os.path.join(self.path, "js", "plugins")
        os.makedirs(self.js_folder)
        with open(os.path.join(self.js_folder, "something.js"), "w") as jsfile:
            jsfile.write("alert(1);\n")

        self.css_folder = os.path.join(self.path, "css", "skins")
        os.makedirs(self.css_folder)
        with open(os.path.join(self.css_folder, "dark.css"), "w") as cssfile:
            cssfile.write("body { background-color: black; }\n")

        with open(os.path.join(self.path, "ignoreme"), "w") as ignorefile:
            ignorefile.write("something to ignore")

    def cleanup(self):
        shutil.rmtree(self.path)



#!/usr/bin/env python

import argparse
import smtplib
import subprocess
from email.mime.text import MIMEText

from ..subcommand import SubCommand
from .reset import Reset


class Record(SubCommand):
    """Record objects from the Data.FS to the file system"""

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "--lasttxn",
            action="store_true",
            default=False,
            help="Add paths mentioned in transactions since the last used",
        )
        parser.add_argument(
            "--commit",
            action="store_true",
            default=False,
            help="Commit changes and send summary mail if there are any",
        )
        parser.add_argument(
            "--autoreset",
            action="store_true",
            default=False,
            help="Automatically reset changes after sending mail for --commit",
        )
        parser.add_argument(
            "--no-recurse",
            action="store_true",
            default=False,
            help="Record only specified paths without recursing",
        )
        parser.add_argument(
            "--skip-errors",
            action="store_true",
            help="Skip failed objects and continue",
            default=False,
        )
        parser.add_argument(
            "path",
            type=str,
            nargs="*",
            help="Sub-Path in Data.fs to be recorded",
        )

    def commit(self):
        """
        Do a commit of all unstaged changes and optionally send an email with a
        summary.
        """
        commit_message = self.config["commit_message"]
        self.gitcmd_run("add", ".")
        try:
            self.gitcmd_run("commit", "-m", commit_message)
        except subprocess.CalledProcessError:
            # Nothing to commit
            return

        # only send a mail if something has changed
        codechg_mail = self.config.get("codechange_mail", False)
        if codechg_mail:  # pragma: no cover
            self.logger.info("Commit was done! Sending mail...")
            pfsystemid = open("/etc/pfsystemid").read().strip()
            pfsystemname = open("/etc/pfsystemname").read().strip()

            status = self.gitcmd_output("show", "--name-status", "HEAD")

            msg = MIMEText(status, "plain", "utf-8")
            msg["Subject"] = "Commit summary on {} ({})".format(
                pfsystemname, pfsystemid
            )

            recipients = codechg_mail.split()
            for recipient in recipients:
                msg["To"] = recipient

            msg["From"] = self.config.get("codechange_sender", "codechanges@perfact.de")

            smtp = smtplib.SMTP("localhost")
            smtp.sendmail(msg["From"], recipients, msg.as_string())
            smtp.quit()

        if self.args.autoreset:
            reset = Reset(sync=self.sync, logger=self.logger, config=self.config)
            parser = argparse.ArgumentParser()
            parser.add_argument("--no-lock", action="store_true")
            reset.add_args(parser)
            reset.args = parser.parse_args(["--no-lock", "HEAD~"])
            reset.run()

    @SubCommand.with_lock
    def run(self):
        paths = self.args.path
        recurse = not self.args.no_recurse

        if self.args.lasttxn:
            # We mean to read from the newest entry
            lasttxn = self.sync.txn_read() or None

            res = self.sync.recent_changes(
                since_secs=None,
                txnid=lasttxn,
                limit=51,
            )
            newest_txnid = res["newest_txnid"]
            if res["search_limit_reached"] or res["limit_reached"] or res["no_records"]:
                # Limits reached mean we need to perform a full dump to
                # recover. The same if there is no transaction present,
                # probably due to a pack of the ZODB.
                paths.append("/")
                recurse = True
            else:
                paths.extend(res["paths"])

        self.sync.record(
            paths=paths, recurse=recurse, skip_errors=self.args.skip_errors
        )

        if self.args.commit:
            self.commit()

        if self.args.lasttxn and (newest_txnid != lasttxn):
            self.sync.txn_write(newest_txnid or "")



#!/usr/bin/python3

import os

from ..helpers import StrRepr, literal_eval
from ..subcommand import SubCommand
from ..zodbsync import mod_format


class Reformat(SubCommand):
    """
    Rewrite commits from given commit to HEAD to post-4.3.2 formatting
    """

    # Note that in contrast to most subcommands, this should probably be
    # executed on a local repository not directly in sync with a ZODB.  No
    # rollback in case of error is implemented yet!

    connect = False

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "commit", type=str, help="Starting point before first commit to rewrite"
        )

    def head(self):
        return self.gitcmd_output("rev-parse", "HEAD").strip()

    @SubCommand.with_lock
    def run(self):
        start = self.args.commit
        commits_raw = self.gitcmd_output(
            "log", "--format=%H", "--reverse", "{}..HEAD".format(start)
        )
        commits = [c for c in commits_raw.strip().split("\n") if c]

        self.gitcmd_run("reset", "--hard", start)
        base = self.config["base_dir"]
        paths = []
        for root, dirs, files in os.walk(os.path.join(base, "__root__")):
            if "__meta__" in files:
                paths.append(os.path.join(root, "__meta__"))
        if self.reformat(paths):
            self.gitcmd_run("commit", "-a", "-m", "zodbsync reformat")
        for idx, commit in enumerate(commits):
            print("Processing commit {}/{}".format(idx + 1, len(commits)))
            cur = self.head()
            paths = list(
                {
                    os.path.join(base, line)
                    for line in self.gitcmd_output(
                        "diff", "--name-only", "--no-renames", commit + "~", commit
                    )
                    .strip()
                    .split("\n")
                    if line
                }
            )
            metas = {path for path in paths if path.endswith("/__meta__")}
            if self.reformat(metas, True):
                self.gitcmd_run("commit", "-a", "-m", "reverse")
            self.gitcmd_run("checkout", "--no-overlay", commit, "--", *paths)
            self.gitcmd_try("commit", "--no-edit", "-c", commit)
            self.reformat(metas)
            # Squash commits together with original message
            self.gitcmd_run("reset", cur)
            while paths:
                self.gitcmd_run("add", *paths[:100])
                del paths[:100]
            self.gitcmd_try("commit", "--no-edit", "-c", commit)

    def reformat(self, paths, legacy=False):
        changed = False
        for path in paths:
            if not os.path.exists(path):
                continue
            for _ in range(1000):  # Just in case we have no stdin
                with open(path) as f:
                    orig = f.read()
                try:
                    data = literal_eval(orig)
                except Exception:
                    # This allows the user to open the file manually, make it
                    # valid to be parsed by literal_eval and continue the
                    # process
                    print("Unable to parse path", path)
                    input("Retrying. Ctrl+C to cancel.")
                    continue
                break
            else:
                raise ValueError("Meta file could not be parsed")
            if legacy:
                fmt = StrRepr()(data, legacy=True)
            else:
                fmt = mod_format(data)
            if orig == fmt:
                continue
            with open(path, "w") as f:
                f.write(fmt)
            changed = True

        return changed



#!/usr/bin/env python
import os

from ..subcommand import SubCommand


class Freeze(SubCommand):
    """Mark paths as frozen and record them"""

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "path",
            type=str,
            nargs="*",
            help="Sub-Path in Data.fs to be frozen",
        )

    @SubCommand.with_lock
    def run(self):
        for path in self.args.path:
            fullpath = self.sync.fs_path(path)
            os.makedirs(fullpath, exist_ok=True)
            with open("{}/__frozen__".format(fullpath), "w"):
                pass
        self.sync.record(paths=self.args.path, recurse=True)



#!/usr/bin/env python

import os
import sys
import warnings

from ..subcommand import SubCommand
from ..zodbsync import mod_format

PY2 = sys.version_info.major == 2


def create_template(type, content_type=None):
    result = {"type": type, "title": ""}
    if content_type is not None:
        result["props"] = [
            [("id", "content_type"), ("type", "string"), ("value", content_type)]
        ]
    return result


META_TEMPLATES = {
    "folder": create_template("Folder"),
    "js": create_template("File", "application/javascript"),
    "css": create_template("File", "text/css"),
    "html": create_template("File", "text/html"),
    "odt": create_template("File", "application/vnd.oasis.opendocument.text"),
    "ods": create_template("File", "application/vnd.oasis.opendocument.spreadsheet"),
    "pdf": create_template("File", "application/pdf"),
    "svg": create_template("File", "image/svg+xml"),
}
META_DEFAULT = create_template("File", "application/octet-stream")


class Upload(SubCommand):
    """[DEPRECATED] Upload a folder structure, e.g. a JS library, to zope
    Data.FS"""

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "source",
            type=str,
            help="Path of source library folder",
        )
        parser.add_argument(
            "path",
            type=str,
            help="Sub-Path in Data.fs to put source folder",
        )
        parser.add_argument(
            "--override",
            "-o",
            action="store_true",
            help="Override object type changes when uploading",
            default=False,
        )
        parser.add_argument(
            "--skip-errors",
            action="store_true",
            help="Skip failed objects and continue",
            default=False,
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Roll back at the end.",
        )
        parser.add_argument(
            "--replace-periods",
            action="store_true",
            default=False,
            help="Replace periods in file names with underscores",
        )
        parser.add_argument(
            "--valid-extensions",
            type=str,
            help=(
                "Only upload files with the extensions listed "
                "(comma separated list). Allow all extensions by default."
            ),
        )

    @SubCommand.with_lock
    def run(self):
        """
        Convert source folder into zodbsync compatible struct in repodir
        and upload it.
        """
        warnings.warn(
            "Static external assets should not be included in the Data.FS",
            DeprecationWarning,
        )
        self.check_repo()

        # we need both filesystem and Data.fs path representation
        data_fs_path, filesystem_path = self.datafs_filesystem_path(self.args.path)

        valid_extensions = self.args.valid_extensions
        if valid_extensions:
            # Parse comma separated list
            valid_extensions = [
                item.strip()
                for item in valid_extensions.split(",")
                if len(item.strip()) > 0
            ]

        # conversion loop: iterate over source folder, create folders in
        # repodir and corresponding files
        for cur_dir_path, dirs, files in os.walk(self.args.source):
            # relative path to be created in repodir
            cur_dir = os.path.relpath(cur_dir_path, self.args.source)

            # repodir folder creation
            new_folder = os.path.join(filesystem_path, cur_dir)
            if PY2:
                os.makedirs(new_folder)
            else:
                os.makedirs(new_folder, exist_ok=True)

            # do not forget meta file for folder
            self.create_file(
                file_path=os.path.join(new_folder, "__meta__"),
                content=mod_format(META_TEMPLATES["folder"]),
            )

            # now check files inside of folder
            for filename in files:
                file_ending = filename.split(".")[-1]

                # bail out if not a valid extension
                if valid_extensions is not None and file_ending not in valid_extensions:
                    continue

                # read file content from source file
                with open(os.path.join(cur_dir_path, filename), "rb") as sourcefile:
                    file_content = sourcefile.read()

                # choose the original filename, or replace periods
                if self.args.replace_periods:
                    repo_filename = filename.replace(".", "_")
                else:
                    repo_filename = filename

                # in repo each file gets its own folder ...
                new_file_folder = os.path.join(new_folder, repo_filename)
                os.makedirs(new_file_folder)

                # ... containing __meta__ and __source__ file
                self.create_file(
                    file_path=os.path.join(new_file_folder, "__meta__"),
                    content=mod_format(META_TEMPLATES.get(file_ending, META_DEFAULT)),
                )
                self.create_file(
                    file_path=os.path.join(
                        new_file_folder, "__source__." + file_ending
                    ),
                    content=file_content,
                    binary=True,
                )

        # conversion done, start playback
        try:
            self.sync.playback_paths(
                paths=[data_fs_path],
                recurse=True,
                override=self.args.override,
                skip_errors=self.args.skip_errors,
                dryrun=self.args.dry_run,
            )

            if self.args.dry_run:
                self.abort()

        except Exception:
            self.logger.exception("Error uploading files. Resetting.")
            self.abort()
            raise



# -*- coding: utf-8 -*-
import ast
import importlib
import operator


class Namespace(object):
    """
    Convert a dict to a namespace, allowing access via a.b instead of a['b']
    """

    def __init__(self, data=None, **kw):
        if data:
            self.__dict__.update(data)
        self.__dict__.update(kw)


def to_string(value, enc="utf-8"):
    """Convert input into a string"""
    if isinstance(value, str):
        return value
    if isinstance(value, bytes):
        return value.decode(enc)
    return str(value)


def to_bytes(value, enc="utf-8"):
    """Convert input to bytes (encoded strings)"""
    if isinstance(value, memoryview):
        return value.tobytes()
    if isinstance(value, bytes):
        return value
    if isinstance(value, str):
        return value.encode(enc)
    return to_bytes(str(value))


def remove_redundant_paths(paths):
    """
    Sort list of paths and remove items that are redundant if remaining
    paths are processed recursively, i.e., if /a/b/ as well as /a/ are
    included, remove /a/b/. Works in-place and also returns the list.
    """
    paths.sort()
    i = 0
    last = None
    while i < len(paths):
        current = paths[i].rstrip("/") + "/"
        if last is not None and current.startswith(last):
            del paths[i]
            continue
        last = current
        i += 1
    return paths


class StrRepr:
    """Create a printable output of the given object data.
    Dicts are converted to sorted lists of tuples, tuples and lists recurse
    into their elements. The top-level element should be a dict.
    `seprules` is a dictionary mapping from keys of the top-level dict to a
    list of levels which should be split into separate lines if they contain an
    iterable, in addition to the default (split the zeroth level and split the
    second one if it is a list).
    `legacy` mode turns off line splitting for iterables with less than two
    items and puts the closing bracket on the same indentation level as the
    items except for the top level.
    """

    def _collect(self, data, level=0, nl="\n"):
        "Internal recursion worker"

        if not isinstance(data, (list, tuple)):
            self.output.append(repr(data))
            return

        # start new line for each element
        linesep = (
            level == 0
            or level == 2
            and isinstance(data, list)
            or level in self.seprules.get(self.section, [])
        )
        if self.legacy and len(data) < 2:
            linesep = False
        # add separator after last element - usually only for lists that are
        # split
        lastsep = linesep

        if isinstance(data, list):
            opn, cls = "[", "]"

        if isinstance(data, tuple):
            opn, cls = "(", ")"
            if len(data) == 1:
                lastsep = True

        self.output.append(opn)
        incnl = nl + "    "
        for idx, item in enumerate(data):
            if level == 0:
                self.section = item[0]
            if linesep:
                self.output.append(incnl)
                self._collect(item, level + 1, incnl)
                self.output.append(",")
            else:
                self._collect(item, level + 1, nl)
                if idx < len(data) - 1 or lastsep:
                    self.output.append(", ")
        if self.legacy and linesep and level > 0:
            self.output.append(incnl + cls)
        else:
            self.output.append(nl + cls if linesep else cls)

    def __call__(self, data, seprules=None, legacy=False):
        "Collect output parts recursively and return their concatenation"
        self.output = []
        self.section = None
        self.seprules = seprules or {}
        self.legacy = legacy

        if isinstance(data, dict):
            data = sorted(data.items())
        self._collect(data)
        return "".join(self.output) + "\n"


def read_pdata(obj):
    """Avoid authentication problems when reading linked pdata."""
    if isinstance(obj.data, (bytes, str)):
        source = obj.data
    else:
        data = obj.data
        if isinstance(data.data, bytes):
            source = b""
        elif isinstance(data.data, str):
            source = ""
        while data is not None:
            source += data.data
            data = data.next
    return source


def literal_eval(value):
    """Literal evaluator (with a bit more power than PT).

    This evaluator is capable of parsing large data sets, and it has
    basic arithmetic operators included.
    """
    if isinstance(value, (bytes, str)):
        value = ast.parse(value, mode="eval")

    bin_ops = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.truediv,
        ast.Mod: operator.mod,
    }

    unary_ops = {
        ast.USub: operator.neg,
    }

    def _convert(node):
        if isinstance(node, ast.Expression):
            return _convert(node.body)
        elif isinstance(node, ast.Constant):
            return node.value
        elif isinstance(node, ast.Tuple):
            return tuple(map(_convert, node.elts))
        elif isinstance(node, ast.List):
            return list(map(_convert, node.elts))
        elif isinstance(node, ast.Dict):
            return dict(
                (_convert(k), _convert(v)) for k, v in zip(node.keys, node.values)
            )
        elif isinstance(node, ast.BinOp):
            return bin_ops[type(node.op)](_convert(node.left), _convert(node.right))
        elif isinstance(node, ast.UnaryOp):
            return unary_ops[type(node.op)](_convert(node.operand))
        raise Exception("Unsupported type {}".format(repr(node)))

    return _convert(value)


def prop_dict(data):
    props = {}

    # Get the properties from object data
    p = dict(data).get("props", None)
    if not p:
        return props

    # Convert each property into a dictionary
    for item in p:
        pd = dict(item)
        # Extract only the value
        props[pd["id"]] = pd["value"]

    return props


def load_config(filename):
    """Load the module at "filename" as module "name". Return the contents
    as a dictionary. Skips contents starting with '_'.
    """
    loader = importlib.machinery.SourceFileLoader("config", filename)
    spec = importlib.util.spec_from_loader(loader.name, loader)
    mod = importlib.util.module_from_spec(spec)
    loader.exec_module(mod)

    return {name: getattr(mod, name) for name in dir(mod) if not name.startswith("_")}


# Helper for handling transaction IDs (which are byte strings of length 8)
def increment_txnid(s):
    """add 1 to s, but for s being a string of bytes"""
    arr = bytearray(s)
    pos = len(arr) - 1
    while pos >= 0:
        if arr[pos] == 255:
            arr[pos] = 0
            pos -= 1
        else:
            arr[pos] += 1
            break
    return bytes(arr)


def obj_modtime(obj):  # pragma: no cover
    """
    Allow access to private method of an object to read out the modtime.
    """
    return obj._p_mtime


def db_modtime(context):  # pragma: no cover
    """
    Allow access to a modtime for the whole ZODB.
    """
    return context._p_jar._db.lastTransaction()


def path_diff(old, new):
    """
    For two lists of tuples (path, checksum) that are ordered by path, return
    the set of all paths that differ, i.e. either are only present in one of
    the lists or have a different checksum.
    """
    result = set()
    oldidx = 0
    newidx = 0
    # Iterate through results, which are ordered by path. Add any
    # deviation to paths
    while oldidx < len(old) and newidx < len(new):
        if old[oldidx] == new[newidx]:
            oldidx += 1
            newidx += 1
            continue
        oldpath = old[oldidx][0]
        newpath = new[newidx][0]
        if oldpath <= newpath:
            result.add(oldpath)
            oldidx += 1
            continue
        if newpath <= oldpath:
            result.add(newpath)
            newidx += 1
    result.update([row[0] for row in old[oldidx:]])
    result.update([row[0] for row in new[newidx:]])
    return result



#!/usr/bin/env python

from ..subcommand import SubCommand


class Pick(SubCommand):
    """Cherry-pick commits, apply them and play back affected objects"""

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "--skip-errors",
            action="store_true",
            default=False,
            help="Skip failed objects and continue",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Only check for conflicts and roll back at the end.",
        )
        parser.add_argument(
            "--grep",
            type=str,
            help="""Find commits starting from the given
            ones, limiting to those with commit messages matching the
            pattern - like "git log --grep".""",
        )
        parser.add_argument(
            "--since",
            type=str,
            help="""Find commits since the given timestamp
             - like "git log --since".""",
        )
        parser.add_argument(
            "--until",
            type=str,
            help="""Find commits until the given timestamp
             - like "git log --until".""",
        )
        parser.add_argument(
            "commit",
            type=str,
            nargs="*",
            help="""Commits that are checked for compatibility and applied,
            playing back all affected paths at the end.""",
        )

    @SubCommand.gitexec
    def run(self):
        commits = []
        if self.args.grep or self.args.since or self.args.until:
            cmd = ["log", "--format=%H", "--reverse"]
            if self.args.grep:
                cmd.extend(["--grep", self.args.grep])
            if self.args.since:
                cmd.extend(["--since", self.args.since])
            if self.args.until:
                cmd.extend(["--until", self.args.until])
            commits = self.gitcmd_output(*cmd, *self.args.commit).split("\n")
        else:
            for commit in self.args.commit:
                if ".." in commit:
                    # commit range
                    commits.extend(
                        self.gitcmd_output(
                            "log", "--format=%H", "--reverse", commit
                        ).split("\n")
                    )
                else:
                    commits.append(commit)

        for commit in commits:
            if not commit:
                continue
            self.logger.info("Checking and applying %s." % commit)
            # capture output and discard so we don't clutter stdout
            # Python 2 has no subprocess.DEVNULL.
            self.gitcmd_output("cherry-pick", "--strategy", "resolve", commit)
